#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: maskbits.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from enum import Enum, Flag, ReprEnum, auto

from strenum import UppercaseStrEnum


__all__ = ["ErrorCode", "Event", "ObserverStageStatus"]


class ErrorCode(Enum):
    """List of error codes."""

    UNCATEGORISED_ERROR = 0
    NOT_IMPLEMENTED = 1
    COMMAND_FAILED = 2
    COMMAND_TIMEDOUT = 3
    USAGE_ERROR = 4
    TIMEOUT = 5
    OVERATCHER_RUNNING = 6
    DEVICE_ERROR = 10
    TELESCOPE_ERROR = 100
    CANNOT_MOVE_LOCAL_MODE = 101
    FAILED_REACHING_COMMANDED_POSITION = 102
    INVALID_TELESCOPE_POSITION = 103
    MOTAN_DEVICE_NOT_REACHABLE = 104
    MOTAN_DEVICE_CANNOT_STOP = 105
    FIBSEL_INVALID_POSITION = 110
    AG_ERROR = 200
    SPECTROGRAPH_ERROR = 300
    SECTROGRAPH_FAILED_EXPOSING = 301
    SECTROGRAPH_NOT_IDLE = 302
    INVALID_CALIBRATION_SEQUENCE = 303
    NPS_ERROR = 400
    ENCLOSURE_ERROR = 500
    ENCLOSURE_IN_LOCAL = 501
    DOOR_STATUS_FAILED = 502
    GUIDER_ERROR = 600
    INVALID_PIXEL_NAME = 601
    SCHEDULER_UNCATEGORISED = 701
    SCHEDULER_TILE_ERROR = 702
    SCHEDULER_CANNOT_FIND_TILE = 703
    OBSERVER_ERROR = 800
    ACQUISITION_FAILED = 801
    CALIBRATION_ERROR = 900
    UNKNOWN_ERROR = 9999

    def is_telescope_error(self):
        """Returns True if the error is related to the telescope."""

        return self.value >= 100 and self.value < 200

    def is_ag_error(self):
        """Returns True if the error is related to the autoguider."""

        return self.value >= 200 and self.value < 300

    def is_spectrograph_error(self):
        """Returns True if the error is related to the spectrograph."""

        return self.value >= 300 and self.value < 400

    def is_nps_error(self):
        """Returns True if the error is related to the NPS."""

        return self.value >= 400 and self.value < 500

    def is_enclosure_error(self):
        """Returns True if the error is related to the enclosure."""

        return self.value >= 500 and self.value < 600

    def is_guiding_error(self):
        """Returns True if the error is related to the guider."""

        return self.value >= 600 and self.value < 700

    def is_scheduler_error(self):
        """Returns True if the error is related to the scheduler."""

        return self.value >= 700 and self.value < 800

    def is_observer_error(self):
        """Returns True if the error is related to the observer."""

        return self.value >= 800 and self.value < 900


class GuiderStatus(Flag):
    """Maskbits with the guider status."""

    IDLE = 1 << 0
    ACQUIRING = 1 << 1
    GUIDING = 1 << 2
    EXPOSING = 1 << 3
    PROCESSING = 1 << 4
    CORRECTING = 1 << 5
    STOPPING = 1 << 6
    FAILED = 1 << 7
    WAITING = 1 << 8
    DRIFTING = 1 << 9

    NON_IDLE = (
        ACQUIRING
        | GUIDING
        | EXPOSING
        | PROCESSING
        | CORRECTING
        | STOPPING
        | WAITING
        | DRIFTING
    )

    def get_names(self):
        """Returns a list of active bit names."""

        return [bit.name for bit in GuiderStatus if self & bit and bit.name]

    def __repr__(self):
        return str(" | ".join(self.get_names()))


class UpperStrEnum(str, ReprEnum):
    """A string enum in which the auto value is the uppercase name."""

    @staticmethod
    def _generate_next_value_(name, *_):
        return name.upper()


class Event(UppercaseStrEnum):
    """Enumeration with the event types."""

    ERROR = auto()
    RECIPE_START = auto()
    RECIPE_END = auto()
    RECIPE_FAILED = auto()
    OBSERVER_NEW_TILE = auto()
    OBSERVER_STAGE_RUNNING = auto()
    OBSERVER_STAGE_DONE = auto()
    OBSERVER_STAGE_FAILED = auto()
    OBSERVER_ACQUISITION_START = auto()
    OBSERVER_ACQUISITION_DONE = auto()
    OBSERVER_STANDARD_ACQUISITION_FAILED = auto()
    DOME_OPENING = auto()
    DOME_OPEN = auto()
    DOME_CLOSING = auto()
    DOME_CLOSED = auto()
    EMERGENCY_SHUTDOWN = auto()
    UNEXPECTED_FIBSEL_REHOME = auto()
    UNCATEGORISED = auto()


class ObserverStageStatus(UpperStrEnum):
    """An enumeration of observer stages."""

    WAITING = auto()
    RUNNING = auto()
    DONE = auto()
    FAILED = auto()
    CANCELLED = auto()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-10
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import concurrent.futures
import datetime
import functools
import hashlib
import logging
import os
import pathlib
import re
import tempfile
import warnings
from contextlib import asynccontextmanager, contextmanager, suppress
from functools import partial, wraps

from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Literal,
    Sequence,
    TypeVar,
)

import httpx
import numpy
import peewee
import polars
import redis
from astropy import units as uu
from astropy.coordinates import angular_separation as astropy_angular_separation
from redis import asyncio as aioredis

from clu import AMQPClient
from sdsstools import get_sjd

from gort import config
from gort.exceptions import ErrorCode, GortError


if TYPE_CHECKING:
    from clu import AMQPClient, AMQPReply

    from gort.devices.telescope import FibSel
    from gort.gort import Gort


__all__ = [
    "get_valid_variable_name",
    "get_next_tile_id",
    "get_calibrators",
    "get_next_tile_id_sync",
    "get_calibrators_sync",
    "register_observation",
    "get_ccd_frame_path",
    "move_mask_interval",
    "angular_separation",
    "get_db_connection",
    "redis_client_async",
    "redis_client_sync",
    "run_in_executor",
    "is_interactive",
    "is_notebook",
    "cancel_task",
    "get_ephemeris_summary",
    "get_ephemeris_summary_sync",
    "get_temporary_file_path",
    "insert_to_database",
    "get_md5sum_file",
    "get_md5sum_from_spectro",
    "get_md5sum",
    "mark_exposure_bad",
    "handle_signals",
    "get_lvmapi_route",
    "run_lvmapi_task",
    "GuiderMonitor",
    "overwatcher_is_running",
    "get_by_source_id",
    "is_actor_running",
    "check_overwatcher_not_running",
    "kubernetes_restart_deployment",
    "kubernetes_list_deployments",
    "get_gort_client",
    "add_night_log_comment",
    "async_noop",
    "LogNamespace",
    "decap",
    "ensure_period",
    "try_or_pass",
]

AnyPath = str | os.PathLike

CAMERAS = [
    "sci.west",
    "sci.east",
    "skye.west",
    "skye.east",
    "skyw.west",
    "skyw.east",
    "spec.east",
]


def get_valid_variable_name(var_name: str):
    """Converts a string to a valid variable name."""

    return re.sub(r"\W|^(?=\d)", "_", var_name)


def get_next_tile_id_sync() -> dict:
    """Retrieves the next ``tile_id`` from the scheduler API. Synchronous version."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    with httpx.Client(base_url=f"http://{host}:{port}/") as client:
        resp = client.get("next_tile")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /next_tile")
        tile_id_data = resp.json()

    return tile_id_data


async def get_next_tile_id() -> dict:
    """Retrieves the next ``tile_id`` from the scheduler API."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient(base_url=f"http://{host}:{port}/") as client:
        resp = await client.get("next_tile")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /next_tile")
        tile_id_data = resp.json()

    return tile_id_data


def get_calibrators_sync(
    tile_id: int | None = None,
    ra: float | None = None,
    dec: float | None = None,
) -> dict:
    """Get calibrators for a ``tile_id`` or science pointing. Synchronous version."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    with httpx.Client(base_url=f"http://{host}:{port}/") as client:
        if tile_id:
            resp = client.get("cals", params={"tile_id": tile_id})
        elif ra is not None and dec is not None:
            resp = client.get("cals", params={"ra": ra, "dec": dec})
        else:
            raise ValueError("ra and dec are required.")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /cals")

    return resp.json()


async def get_calibrators(
    tile_id: int | None = None,
    ra: float | None = None,
    dec: float | None = None,
):
    """Get calibrators for a ``tile_id`` or science pointing."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient(base_url=f"http://{host}:{port}/") as client:
        if tile_id:
            resp = await client.get("cals", params={"tile_id": tile_id})
        elif ra is not None and dec is not None:
            resp = await client.get("cals", params={"ra": ra, "dec": dec})
        else:
            raise ValueError("ra and dec are required.")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /cals")

    return resp.json()


async def register_observation(payload: dict):
    """Registers an observation with the scheduler."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient() as client:
        resp = await client.put(
            f"http://{host}:{port}/register_observation",
            json=payload,
            follow_redirects=True,
        )

        if resp.status_code != 200 or not resp.json()["success"]:
            raise RuntimeError(f"Failed registering observation: {resp.text}.")


def mark_exposure_bad(tile_id: int, dither_position: int = 0):
    """Marks a registered tile/dither as bad."""

    db = get_db_connection()

    completion_status = peewee.Table("completion_status", schema="lvmopsdb").bind(db)
    dither = peewee.Table("dither", schema="lvmopsdb").bind(db)

    dither_pk = (
        dither.select(dither.c.pk)
        .where(
            dither.c.tile_id == tile_id,
            dither.c.position == dither_position,
        )
        .namedtuples()
    )

    if len(dither_pk) == 0:
        raise ValueError("No matching tile-position.")

    completion_status.update(done=False).where(
        completion_status.c.pk == dither_pk[0].pk
    ).execute()


async def set_tile_status(tile_id: int, note: str | None = None):
    """Enables/disables a tile in the database."""

    sch_config = config["services"]["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient() as client:
        resp = await client.put(
            f"http://{host}:{port}/tile_status",
            params={"tile_id": tile_id, "note": note},
            json={},
            follow_redirects=True,
        )

        if resp.status_code != 200 or not resp.json()["success"]:
            raise RuntimeError(f"Failed setting tile status: {resp.text}.")


def is_notebook() -> bool:
    """Returns :obj:`True` if the code is run inside a Jupyter Notebook.

    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    """

    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def is_interactive():
    """Returns :obj:`True` is we are in an interactive session."""

    import __main__ as main

    return not hasattr(main, "__file__")


def get_ccd_frame_path(
    frame_id: int,
    sjd: int | None = None,
    cameras: str | list[str] | None = None,
    spectro_path="/data/spectro",
) -> list[str]:
    """Returns the paths for the files for a spectrograph frame.

    Parameters
    ----------
    frame_id
        The spectrograph frame for which the paths are searched.
    mjd
        The SJD in which the frames where taken. If not provided, all the
        directories under ``spectro_path`` are searched.
    cameras
        The cameras to be returned. If :obj:`None`, all cameras found are returned.
    spectro_path
        The path to the ``spectro`` directory where spectrograph files are
        stored under an SJD structure.

    Returns
    -------
    paths
        The list of paths to CCD frames that match ``frame_id``.

    """

    if isinstance(cameras, str):
        cameras = [cameras]

    base_path = pathlib.Path(spectro_path)
    recursive = True
    if sjd:
        base_path /= str(sjd)
        recursive = False

    # Glob all files that match the frame_id.
    globp = f"*{frame_id}.fits.*"
    if recursive:
        globp = f"**/{globp}"

    files = [str(path) for path in base_path.glob(globp)]

    if cameras is None:
        return files

    files_camera = []
    for camera in cameras:
        for file in files:
            if f"-{camera}-" in file:
                files_camera.append(file)

    return files_camera


async def move_mask_interval(
    gort: Gort,
    positions: str | list[str] = "P1-*",
    order_by_steps: bool = False,
    total_time: float | None = None,
    time_per_position: float | None = None,
    notifier: Callable[[str], None] | Callable[[str], Coroutine] | None = None,
):
    """Moves the fibre mask in the spectrophotometric telescope at intervals.

    Parameters
    ----------
    gort
        The instance of :obj:`.Gort` to communicate with the actor system.
    positions
        The positions to iterate over. It can be a string in which case it will
        be treated as a regular expression and any mask position that matches the
        value will be iterated, in alphabetic order. Alternative it can be a list
        of positions to move to which will be executed in that order.
    order_by_steps
        If :obj:`True`, the positions are iterated in order of smaller to larger
        number of step motors.
    total_time
        The total time to spend iterating over positions, in seconds. Each position
        will  be visited for an equal amount of time. The time required to move the
        mask will not be taken into account, which means the total execution
        time will be longer than ``total_time``.
    time_per_position
        The time to spend on each mask position, in seconds. The total execution
        time will be ``len(positions)*total_time+overhead`` where ``overhead`` is
        the time required to move the mask between positions.
    notifier
        A function or coroutine to call every time a new position is reached.
        If it's a coroutine, it is scheduled as a task. If it is a normal
        callback it should run quickly to not perceptibly affect the total
        execution time.

    """

    try:
        fibsel: FibSel = gort.telescopes.spec.fibsel
    except Exception as err:
        raise RuntimeError(f"Cannot find fibre selector: {err}")

    if total_time is not None and time_per_position is not None:
        raise ValueError("Only one of total_time or time_per_position can be used.")

    if total_time is None and time_per_position is None:
        raise ValueError("One of total_time or time_per_position needs to be passed.")

    mask_config = gort.config["telescopes"]["mask_positions"]
    all_positions = list(mask_config)

    if isinstance(positions, str):
        regex = positions
        all_positions = fibsel.list_positions()
        positions = [pos for pos in all_positions if re.match(regex, pos)]

        if order_by_steps:
            positions = sorted(positions, key=lambda p: mask_config[p])

    fibsel.write_to_log(f"Iterating over positions {positions}.")

    if total_time:
        time_per_position = total_time / len(positions)

    assert time_per_position is not None

    for position in positions:
        await fibsel.move_to_position(position)

        # Notify.
        if notifier is not None:
            if asyncio.iscoroutinefunction(notifier):
                asyncio.create_task(notifier(position))
            else:
                notifier(position)

        await asyncio.sleep(time_per_position)


def angular_separation(lon1: float, lat1: float, lon2: float, lat2: float):
    """A wrapper around astropy's ``angular_separation``.

    Returns the separation between two sets of coordinates. All units must
    be degrees and the returned values is also the separation in degrees.

    """

    separation = astropy_angular_separation(
        lon1 * uu.degree,  # type: ignore
        lat1 * uu.degree,  # type: ignore
        lon2 * uu.degree,  # type: ignore
        lat2 * uu.degree,  # type: ignore
    )

    return separation.to("deg").value


def get_db_connection():
    """Returns a DB connection from the configuration file parameters."""

    conn = peewee.PostgresqlDatabase(**config["services"]["database"]["connection"])
    assert conn.connect(), "Database connection failed."

    return conn


@asynccontextmanager
async def redis_client_async() -> AsyncGenerator[aioredis.Redis, None]:
    """Returns a Redis connection from the configuration file parameters."""

    redis_config = config["services"]["redis"]
    client = aioredis.from_url(redis_config["url"], decode_responses=True)

    yield client

    await client.aclose()


@contextmanager
def redis_client_sync() -> Generator[redis.Redis, None]:
    """Returns a Redis connection from the configuration file parameters."""

    redis_config = config["services"]["redis"]
    client = redis.from_url(redis_config["url"], decode_responses=True)

    yield client

    client.close()


async def run_in_executor(fn, *args, catch_warnings=False, executor="thread", **kwargs):
    """Runs a function in an executor.

    In addition to streamlining the use of the executor, this function
    catches any warning issued during the execution and reissues them
    after the executor is done. This is important when using the
    actor log handler since inside the executor there is no loop that
    CLU can use to output the warnings.

    In general, note that the function must not try to do anything with
    the actor since they run on different loops.

    """

    fn = partial(fn, *args, **kwargs)

    if executor == "thread":
        executor = concurrent.futures.ThreadPoolExecutor
    elif executor == "process":
        executor = concurrent.futures.ProcessPoolExecutor
    else:
        raise ValueError("Invalid executor name.")

    if catch_warnings:
        with warnings.catch_warnings(record=True) as records:
            with executor() as pool:
                result = await asyncio.get_event_loop().run_in_executor(pool, fn)

        for ww in records:
            warnings.warn(ww.message, ww.category)

    else:
        with executor() as pool:
            result = await asyncio.get_running_loop().run_in_executor(pool, fn)

    return result


async def cancel_task(task: asyncio.Future | None):
    """Safely cancels a task."""

    if task is None or task.done():
        return

    task.cancel()
    with suppress(asyncio.CancelledError):
        await task


def get_temporary_file_path(*args, create_parents: bool = False, **kwargs):
    """Returns a valid path to a temporary file.

    `args` and `kwargs` are directly passed to `tempfile.NamedTemporaryFile`.
    If `create_parents`, the parent directories are created if they don't
    exist.

    """

    tmp_log_file = tempfile.NamedTemporaryFile(*args, **kwargs)
    tmp_log_file.close()

    tmp_path = pathlib.Path(tmp_log_file.name)
    if tmp_path.exists():
        tmp_path.unlink()
    elif create_parents:
        tmp_path.parent.mkdir(parents=True, exist_ok=True)

    return tmp_path


def insert_to_database(
    table_name: str,
    payload: list[dict[str, Any]],
    columns: list[str] | None = None,
):
    """Inserts data into the database.

    Parameters
    ----------
    table_name
        The table in the database where to insert the data. Can be in the format
        ``schema.table_name``.
    payload
        The data to ingest, as a list of dictionaries in which each dictionary
        is a mapping of column name in ``table`` to the value to ingest.
    columns
        A list of table columns. If not passed, the column names are inferred from
        the first element in the payload. In this case you must ensure that all the
        elements in the payload contain entries for all the columns (use :obj:`None`
        to fill missing data).

    """

    if len(payload) == 0:
        return

    columns = columns or list(payload[0].keys())

    conn = get_db_connection()

    schema: str | None
    if "." in table_name:
        schema, table_name = table_name.split(".")
    else:
        schema = None
        table_name = table_name

    table = peewee.Table(table_name, schema=schema, columns=columns)
    table.bind(conn)

    table.insert(payload).execute()


def get_md5sum_file(file: AnyPath):
    """Returns the path to the MD5 file for the spectro files."""

    file = pathlib.Path(file).absolute()
    mjd = file.parts[-2]

    md5sum = file.parent / f"{mjd}.md5sum"

    return md5sum if md5sum.exists() else None


def get_md5sum_from_spectro(file: AnyPath):
    """Returns the MD5 checksum for a file from the spectro checksum file."""

    file = pathlib.Path(file).absolute()
    basename = file.name

    md5sum_file = get_md5sum_file(file)
    if not md5sum_file:
        return None

    data = open(md5sum_file).read()

    match = re.search(rf"([0-9a-f]+)\s+{basename}", data)
    if not match:
        return None

    return match.groups(1)[0]


def get_md5sum(file: AnyPath):
    """Returns the MD5 checksum for a file."""

    data = open(file, "rb").read()

    return hashlib.md5(data).hexdigest()


def handle_signals(
    signals: Sequence[int],
    callback: Callable[[], Any] | None = None,
    cancel: bool = True,
):
    """Runs a callback when a signal is received during the execution of a task.

    This function is meant to decorate coroutines. If a signal matching ``signals``
    is received, the callback is run and the coroutine (which is executed as a task)
    is cancelled if ``cancel=True``.

    """

    def _handle_signal(task: asyncio.Task):
        if cancel:
            task.cancel()

        if callback:
            asyncio.get_running_loop().call_soon(callback)

    def _outter_wrapper(coro_func):
        @functools.wraps(coro_func)
        async def _inner_wrapper(*args, **kwargs):
            task = asyncio.create_task(coro_func(*args, **kwargs))

            handler = partial(_handle_signal, task)
            for sgn in signals:
                asyncio.get_running_loop().add_signal_handler(sgn, handler)

            try:
                try:
                    return await task
                except asyncio.CancelledError:
                    raise KeyboardInterrupt()
            finally:
                for sgn in signals:
                    asyncio.get_running_loop().remove_signal_handler(sgn)

        return _inner_wrapper

    return _outter_wrapper


def get_by_source_id(source_id: int) -> dict | None:
    """Returns Gaia DR3 information for a source ID."""

    db = get_db_connection()

    gaia_dr3 = peewee.Table("gaia_dr3_source", schema="catalogdb").bind(db)

    query = gaia_dr3.select(gaia_dr3.star).where(gaia_dr3.c.source_id == source_id)
    data = query.dicts()

    if len(data) == 0:
        return None

    return data[0]


async def get_ephemeris_summary(sjd: int | None = None) -> dict:
    """Returns the ephemeris summary from ``lvmapi``."""

    host = config["services"]["lvmapi"]["host"]
    port = config["services"]["lvmapi"]["port"]
    url = f"http://{host}:{port}/ephemeris/summary"

    sjd = sjd or get_sjd("LCO")

    async with httpx.AsyncClient() as client:
        resp = await client.get(url, params={"sjd": sjd})
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /ephemeris/summary")

    return resp.json()


def get_ephemeris_summary_sync(sjd: int | None = None) -> dict:
    """Returns the ephemeris summary from ``lvmapi``."""

    host = config["services"]["lvmapi"]["host"]
    port = config["services"]["lvmapi"]["port"]
    url = f"http://{host}:{port}/ephemeris/summary"

    sjd = sjd or get_sjd("LCO")

    with httpx.Client() as client:
        resp = client.get(url, params={"sjd": sjd})
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /ephemeris/summary")

    return resp.json()


async def get_lvmapi_route(route: str, params: dict = {}, timeout: float = 5, **kwargs):
    """Gets an ``lvmapi`` route."""

    params.update(kwargs)

    host, port = config["services"]["lvmapi"].values()

    if "/" in host:
        host, base_route = host.split("/")
        if base_route != "":
            base_route += "/"
    else:
        base_route = ""

    if route.startswith("/"):
        route = route[1:]

    if "?" in route:
        raise ValueError("Query parameters should be passed as a dictionary.")

    async with httpx.AsyncClient(
        base_url=f"http://{host}:{port}/{base_route}",
        follow_redirects=True,
        timeout=timeout,
    ) as client:
        response = await client.get(route, params=params)

        if (code := response.status_code) != 200:
            raise ValueError(f"Route /{route} failed with error {code}.")

    return response.json()


async def run_lvmapi_task(
    route: str,
    params: dict = {},
    wait: bool = True,
    check_interval: float = 3,
    timeout: float | None = None,
    **kwargs,
):
    """Runs an LVM API task route.

    Task routes immediately start the task in the background and return a task ID.
    The API can then be queried until the task is completed.

    Parameters
    ----------
    route
        The route to the task.
    params
        The parameters to pass to the route.
    wait
        If :obj:`True`, waits until the task is completed and returns the result.
    check_interval
        The interval at which to check the task status. Ignored if ``wait=False``.
    timeout
        The timeout for the task execution. Ignored if ``wait=False``.

    Returns
    -------
    result
        The task ID if ``wait=False`` or the result of the task if ``wait=True``.

    """

    task_id = await get_lvmapi_route(route, params=params, **kwargs)

    if not wait:
        return task_id

    elapsed: float = 0
    while True:
        await asyncio.sleep(check_interval)
        elapsed += check_interval

        ready = await get_lvmapi_route(f"/tasks/{task_id}/ready")
        if ready is True:
            return await get_lvmapi_route(f"/tasks/{task_id}/result")

        if timeout and elapsed > timeout:
            raise asyncio.TimeoutError(f"Task {task_id} did not complete in time.")


class GuiderMonitor:
    """A tool to monitor guider outputs and store them in a dataframe."""

    def __init__(self, gort: Gort, actor: str | None = None):
        self.gort = gort
        self.actor = actor

        self._data: dict[tuple[int, str], dict[str, Any]] = {}

        self._schema = {
            "frameno": polars.Int32(),
            "telescope": polars.String(),
            "time": polars.Datetime(time_zone="UTC"),
            "n_sources": polars.Int32(),
            "focus_position": polars.Float32(),
            "fwhm": polars.Float32(),
            "ra": polars.Float64(),
            "dec": polars.Float64(),
            "ra_offset": polars.Float32(),
            "dec_offset": polars.Float32(),
            "separation": polars.Float32(),
            "pa": polars.Float32(),
            "pa_offset": polars.Float32(),
            "zero_point": polars.Float32(),
            "mode": polars.String(),
            "ax0_applied": polars.Float32(),
            "ax1_applied": polars.Float32(),
            "rot_applied": polars.Float32(),
        }

    def reset(self):
        """Resets the internal state."""

        self.__init__(self.gort, self.actor)

    def start_monitoring(self):
        """Starts monitoring the guider outputs."""

        if self._handle_guider_reply not in self.gort._callbacks:
            self.gort.add_reply_callback(self._handle_guider_reply)

        self.reset()

    def stop_monitoring(self):
        """Stops monitoring the guider outputs."""

        if self._handle_guider_reply in self.gort._callbacks:
            self.gort.remove_reply_callback(self._handle_guider_reply)

    def __del__(self):
        self.stop_monitoring()

    def get_dataframe(self) -> polars.DataFrame | None:
        """Returns the collected data as a dataframe."""

        if len(self._data) == 0:
            return

        values = self._data.values()
        dfs = [polars.DataFrame([dv], schema=self._schema) for dv in values]

        return polars.concat(dfs).sort(["frameno", "telescope"])

    async def _handle_guider_reply(self, reply: AMQPReply):
        """Processes an actor reply and stores the collected data."""

        if self.actor is not None:
            if self.actor not in str(reply.sender):
                return
        else:
            if ".guider" not in str(reply.sender):
                return

        body = reply.body

        telescope = str(reply.sender).split(".")[1]
        frameno: int | None = None
        new_data: dict[str, Any] = {}

        try:
            if "frame" in body:
                frame = body["frame"]
                frameno = frame["seqno"]
                new_data = {
                    "frameno": frameno,
                    "telescope": telescope,
                    "time": datetime.datetime.now(datetime.timezone.utc),
                    "n_sources": frame["n_sources"],
                    "focus_position": frame["focus_position"],
                    "fwhm": frame["fwhm"],
                }

            elif "measured_pointing" in body:
                measured_pointing = body["measured_pointing"]
                frameno = measured_pointing["frameno"]
                new_data = {
                    "frameno": frameno,
                    "telescope": telescope,
                    "ra": measured_pointing["ra"],
                    "dec": measured_pointing["dec"],
                    "ra_offset": measured_pointing["radec_offset"][0],
                    "dec_offset": measured_pointing["radec_offset"][1],
                    "separation": measured_pointing["separation"],
                    "pa": measured_pointing.get("pa", numpy.nan),
                    "pa_offset": measured_pointing.get("pa_offset", numpy.nan),
                    "zero_point": measured_pointing.get("zero_point", numpy.nan),
                    "mode": measured_pointing["mode"],
                }

            elif "correction_applied" in body:
                correction_applied = body["correction_applied"]
                frameno = correction_applied["frameno"]
                new_data = {
                    "frameno": frameno,
                    "telescope": telescope,
                    "ax0_applied": correction_applied["motax_applied"][0],
                    "ax1_applied": correction_applied["motax_applied"][1],
                    "rot_applied": correction_applied.get("rot_applied", 0.0),
                }
            else:
                return

            if not isinstance(frameno, int):
                return

            key = (frameno, telescope)
            if key in self._data:
                self._data[key].update(new_data)
            else:
                self._data[key] = new_data

        except Exception as err:
            self.gort.log.warning(f"Error processing guider reply: {err}")

    def to_header(self):
        """Returns a header with pointing and guiding information."""

        header: dict[str, Any] = {}

        telescopes = ["sci", "spec", "skye", "skyw"]
        if self.actor is not None:
            telescopes = [self.actor.split(".")[1]]

        df = self.get_dataframe()

        if df is not None:
            for tel in telescopes:
                try:
                    tel_data = df.filter(polars.col.telescope == tel)
                    if len(tel_data) < 2:
                        frame0 = None
                        framen = None
                    else:
                        frame0 = int(tel_data["frameno"].min())  # type: ignore
                        framen = int(tel_data["frameno"].max())  # type: ignore

                    header.update(
                        {
                            f"G{tel.upper()}FR0": (frame0, f"{tel} first guider frame"),
                            f"G{tel.upper()}FRN": (framen, f"{tel} last guider frame"),
                        }
                    )

                except Exception as err:
                    self.gort.specs.write_to_log(
                        f"Failed updating guider header information for {tel}: {err}",
                        "warning",
                    )
                    continue

        return header


async def is_actor_running(actor: str, client: AMQPClient | None = None):
    """Returns :obj:`True` if the actor is running.

    This assumes that the actor implements a `ping` command.

    """

    close_client: bool = False
    if not client:
        client = await AMQPClient().start()
        close_client = True

    # TODO: this can fail if the event loop has closed the client connection.
    ping = await client.send_command(actor, "ping")

    if close_client:
        await client.stop()

    if ping.status.did_succeed:
        return True

    return False


async def overwatcher_is_running(client: AMQPClient | None = None):
    """Returns :obj:`True` if the overwatcher is running."""

    overwatcher_actor_name = config["overwatcher"]["actor"]["name"]
    return await is_actor_running(overwatcher_actor_name, client=client)


async def overwatcher_is_enabled(client: AMQPClient | None = None):
    """Returns :obj:`True` if the overwatcher is enabled."""

    close_client: bool = False
    if not client:
        client = await AMQPClient().start()
        close_client = True

    if not (await overwatcher_is_running(client)):
        return False

    overwatcher_actor_name = config["overwatcher"]["actor"]["name"]
    status = await client.send_command(overwatcher_actor_name, "status")

    if close_client:
        await client.stop()

    if status.status.did_succeed:
        return status.replies.get("status")["enabled"]

    raise GortError("Cannot determine if the Overwatcher is running.")


def check_overwatcher_not_running(coro):
    """Decorator that fails a coroutine if the overwatcher is running."""

    @wraps(coro)
    async def wrapper(*args, **kwargs):
        if await overwatcher_is_running():
            raise GortError(
                f"Overwatcher is running. Cannot execute {coro.__name__}.",
                ErrorCode.OVERATCHER_RUNNING,
            )
        return await coro(*args, **kwargs)

    return wrapper


async def kubernetes_list_deployments():
    """Retrieves the Kubernetes deployments from the API."""

    return await get_lvmapi_route("/kubernetes/deployments/list")


async def kubernetes_restart_deployment(name: str):
    """Restarts a Kubernetes deployment via the API."""

    return await get_lvmapi_route(f"/kubernetes/deployments/{name}/restart")


@asynccontextmanager
async def get_gort_client(override_overwatcher: bool | None = None):
    """Returns a GORT client."""

    from gort import Gort

    override_envvar = os.environ.get("GORT_OVERRIDE_OVERWATCHER", "0")
    if override_overwatcher is None:
        override_overwatcher = False if override_envvar == "0" else True

    gort = await Gort(
        override_overwatcher=override_overwatcher,
        verbosity="debug",
    ).init()

    yield gort

    await gort.stop()


NightLogCategories = Literal["weather", "issues", "other", "observers", "overwatcher"]


async def add_night_log_comment(comment: str, category: NightLogCategories = "other"):
    """Adds a comment to the night log."""

    payload = {
        "mjd": get_sjd("LCO"),
        "category": category or "other",
        "comment": comment,
    }

    host, port = config["services"]["lvmapi"].values()

    async with httpx.AsyncClient(
        base_url=f"http://{host}:{port}",
        follow_redirects=True,
    ) as client:
        response = await client.post("/logs/night-logs/comments/add", json=payload)

        code = response.status_code
        if code != 200:
            raise ValueError(f"Failed adding night log comment. Code {code}.")


async def async_noop(*args, **kwargs):
    """A no-op coroutine."""

    return None


class LogNamespace:
    """A namespace for log messages."""

    def __init__(self, logger: logging.Logger, header: str = "") -> None:
        self.logger = logger
        self.header = header

    def debug(self, message: str, *args, **kwargs):
        self.logger.log(logging.DEBUG, self._get_message(message), *args, **kwargs)

    def info(self, message: str, *args, **kwargs):
        self.logger.log(logging.INFO, self._get_message(message), *args, **kwargs)

    def warning(self, message: str, *args, **kwargs):
        self.logger.log(logging.WARNING, self._get_message(message), *args, **kwargs)

    def error(self, message: str, *args, **kwargs):
        self.logger.log(logging.ERROR, self._get_message(message), *args, **kwargs)

    def critical(self, message: str, *args, **kwargs):
        self.logger.log(logging.CRITICAL, self._get_message(message), *args, **kwargs)

    def exception(self, *args, **kwargs):
        self.logger.exception(*args, **kwargs)

    def _get_message(self, message: str):
        return self.header.format(**locals()) + message


def decap(string: Any, add_period: bool = False):
    """Decapitalises the first letter of a string."""

    if not isinstance(string, str):
        string = str(string)

    if len(string) == 0:
        return ""

    decapped = str(string[0].lower() + string[1:])

    if not add_period:
        return decapped

    return ensure_period(decapped)


def ensure_period(string: str):
    """Ensures that a string ends with a period."""

    if not string.strip().endswith("."):
        string += "."

    return string


T = TypeVar("T", bound=Any)


async def try_or_pass(coro: Awaitable[T]) -> T | None:
    """Decorator that wraps a coroutine in a try/except block."""

    try:
        return await coro
    except Exception:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-05
# @Filename: recipes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio

from typing import TYPE_CHECKING, ClassVar

from lvmopstools.utils import is_host_up

from gort.enums import ErrorCode
from gort.exceptions import TroubleshooterCriticalError
from gort.tools import decap, run_lvmapi_task, set_tile_status


if TYPE_CHECKING:
    from .troubleshooter import TroubleModel, Troubleshooter


__all__ = ["TroubleshooterRecipe"]


class TroubleshooterRecipe(metaclass=abc.ABCMeta):
    """Base class for troubleshooter recipes."""

    priority: ClassVar[int | None] = None
    name: ClassVar[str]

    def __init__(self, troubleshooter: Troubleshooter):
        self.troubleshooter = troubleshooter

        self.overwatcher = troubleshooter.overwatcher
        self.gort = troubleshooter.overwatcher.gort

        self.notify = troubleshooter.overwatcher.notify

        assert hasattr(self, "name"), "Recipe must have a name."

    @abc.abstractmethod
    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        raise NotImplementedError

    async def handle(self, error_model: TroubleModel | None = None) -> bool:
        """Runs the recipe to handle the error."""

        try:
            return await self._handle_internal(error_model=error_model)
        except TroubleshooterCriticalError:
            # Propagate this error, which will be handled by the Troubleshooter class.
            raise
        except Exception as err:
            await self.notify(
                f"Error running recipe {self.name}: {decap(err)}",
                level="error",
            )
            return False

    @abc.abstractmethod
    async def _handle_internal(self, error_model: TroubleModel | None = None) -> bool:
        """Internal implementation of the recipe."""

        raise NotImplementedError


class CleanupRecipe(TroubleshooterRecipe):
    """A recipe that cleans up after an error."""

    priority = 500
    name = "cleanup"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        # Always return False. This is a last resort recipe.
        return False

    async def _handle_internal(self, error_model: TroubleModel | None = None) -> bool:
        """Run the cleanup recipe."""

        await self.overwatcher.gort.cleanup(readout=False)

        return True


class AcquisitionFailedRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "acquisition_failed"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code == ErrorCode.ACQUISITION_FAILED:
            return True

        return False

    def get_camera_ips(self):
        """Returns the IPs of the AG cameras."""

        IPs: dict[str, str] = {}
        for ag in self.gort.ags.values():
            for cam, ip in ag.ips.items():
                if ip is not None:
                    IPs[f"{ag.name}-{cam}"] = ip

        return IPs

    async def ping_ag_cameras(self):
        """Pings the AG cameras to see if they are all up."""

        IPs = self.get_camera_ips()
        pings = await asyncio.gather(*[is_host_up(IP) for IP in IPs.values()])

        return {cam_name: pings[ii] for ii, cam_name in enumerate(IPs)}

    async def _handle_disconnected_cameras(self):
        """Handle disconnected cameras."""

        # First check if all the cameras are connected.
        # If not, for now we don't have an automatic way to recover.
        pings = await self.ping_ag_cameras()
        if all(pings):
            await self.notify("All AG cameras ping. Reconnecting cameras.")

            # Reconnect the cameras.
            await self.gort.ags.reconnect()

            # Refresh the remote actors (should not be necessary).
            for ag in self.gort.ags.values():
                await ag.actor.refresh()

            return False  # False means the error was probably not handled

        # Create a list of failed cameras. The format is CAM-<last octet>.
        IPs = self.get_camera_ips()
        failed_cameras: list[str] = []
        for camera, ping in pings.items():
            if not ping:
                last8 = IPs[camera].split(".")[-1]
                failed_cameras.append(f"CAM-{last8}")

        # Power cycle the switch ports to those cameras.
        await self.notify(
            f"Found {len(failed_cameras)} AG cameras that are down: {failed_cameras}. "
            "Power cycling switch ports. This will take several minutes.",
            level="warning",
        )

        await run_lvmapi_task(
            "/macros/power_cycle_ag_cameras",
            params={"cameras": failed_cameras},
        )
        await asyncio.sleep(30)

        pings = await self.ping_ag_cameras()
        if not all(pings):
            raise TroubleshooterCriticalError("Unable to reconnect AG cameras.")

        await self.notify("AG cameras have been power cycled and are now pinging.")

        return True

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        error = error_model.error

        # First check if all the cameras are connected and pinging.
        if await self._handle_disconnected_cameras():
            return True

        # If that didn't work, try to disable the tile. However, if this error has
        # happened several times, there must be some other problem so we shut down.
        if error_model.tracking_data and error_model.tracking_data["count"] >= 3:
            raise TroubleshooterCriticalError(
                "Acquisition failed multiple times. "
                "Finishing the observe loop and disabling the overwatcher."
            )

        tile_id: int | None = error.payload.get("tile_id", None)
        if tile_id is None:
            await self.notify(
                'Cannot disable tile without a "tile_id. '
                "Continuing observations without disabling tile.",
                level="error",
            )
        else:
            await set_tile_status(
                tile_id,
                note=f"Acquisition failed: {decap(error_model.message)}",
            )
            await self.notify(
                f"tile_id={tile_id} has been disabled. Continuing observations.",
                level="warning",
            )

        return True


class SchedulerFailedRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "scheduler_failed"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code.is_scheduler_error():
            return True

        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        await self.notify(
            "The scheduler was not able to find a valid tile to "
            "observe. Waiting 60 seconds before trying again.",
            level="warning",
        )
        await asyncio.sleep(60)

        return True


class SpectrographNotIdleRecipe(TroubleshooterRecipe):
    """Handle acquisition failures."""

    priority = 1
    name = "spectrograph_not_idle"

    def match(self, error_model: TroubleModel) -> bool:
        """Returns True if the recipe can handle the error."""

        if error_model.error_code == ErrorCode.SECTROGRAPH_NOT_IDLE:
            return True

        return False

    async def _handle_internal(self, error_model: TroubleModel) -> bool:
        """Handle the error."""

        await self.notify("Resetting spectrographs.", level="warning")
        await self.gort.specs.reset()

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-05
# @Filename: commands.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import math
import time

from typing import TYPE_CHECKING, Any

import click

from clu.parsers.click import command_parser as overwatcher_cli

from gort.overwatcher.calibrations import CalibrationState


if TYPE_CHECKING:
    from gort.overwatcher.actor.actor import OverwatcherCommand


@overwatcher_cli.command()
async def status(command: OverwatcherCommand):
    """Reports the status of the overwatcher."""

    overwatcher = command.actor.overwatcher

    return command.finish(
        message={
            "status": {
                "sjd": overwatcher.ephemeris.sjd,
                "running": True,
                "enabled": overwatcher.state.enabled,
                "observing": overwatcher.state.observing,
                "calibrating": overwatcher.state.calibrating,
                "safe": overwatcher.state.safe,
                "alerts": overwatcher.state.alerts,
                "night": overwatcher.state.night,
                "allow_calibrations": overwatcher.state.allow_calibrations,
                "dry_run": overwatcher.state.dry_run,
            }
        }
    )


@overwatcher_cli.command()
async def enable(command: OverwatcherCommand):
    """Enables the overwatcher."""

    overwatcher = command.actor.overwatcher

    if not overwatcher.state.enabled:
        overwatcher.state.enabled = True
        await overwatcher.notify("Overwatcher has been enabled.")

    return command.finish()


@overwatcher_cli.command()
@click.option(
    "--now",
    is_flag=True,
    help="Stops observing immediately.",
)
@click.option(
    "--close-dome",
    is_flag=True,
    help="Closes the dome after stopping observing.",
)
async def disable(
    command: OverwatcherCommand,
    now: bool = False,
    close_dome: bool = False,
):
    """Disables the overwatcher."""

    overwatcher = command.actor.overwatcher

    if close_dome and not now:
        return command.fail("--now is required when using --close.")

    if now:
        await overwatcher.shutdown(
            "user disabled the Overwatcher.",
            close_dome=close_dome,
            disable_overwatcher=True,
        )
        return

    if overwatcher.state.enabled:
        # This will cancel the observing loop after the current tile.
        overwatcher.state.enabled = False
        await overwatcher.notify("Overwatcher has been disabled.")

    return command.finish()


@overwatcher_cli.command()
async def reset(command: OverwatcherCommand):
    """Resets the Overwatcher."""

    overwatcher = command.actor.overwatcher

    overwatcher.dome.reset()
    for actor in overwatcher.gort.actors.values():
        await actor.refresh()

    return command.finish()


@overwatcher_cli.group()
def calibrations(*_):
    """Handles the automated calibrations."""

    pass


@calibrations.command()
async def enable_calibrations(command: OverwatcherCommand):
    """Allows dome calibrations."""

    overwatcher = command.actor.overwatcher

    if not overwatcher.state.allow_calibrations:
        overwatcher.state.allow_calibrations = True
        await overwatcher.notify("Calibrations have been enabled.")

    return command.finish()


@calibrations.command()
async def disable_calibrations(command: OverwatcherCommand):
    """Disallows calibrations."""

    overwatcher = command.actor.overwatcher

    if overwatcher.state.allow_calibrations:
        overwatcher.state.allow_calibrations = False
        await overwatcher.notify("Calibrations have been disabled.")

    return command.finish()


@calibrations.command(name="list")
async def list_(command: OverwatcherCommand):
    """Lists the calibrations."""

    def format_timestamp(timestamp: float | None) -> str | None:
        if timestamp is None:
            return None

        return time.strftime("%H:%M:%S", time.gmtime(timestamp))

    cals_overwatcher = command.actor.overwatcher.calibrations
    schedule = cals_overwatcher.schedule
    calibrations = schedule.calibrations

    now = time.time()

    response: list[dict[str, Any]] = []
    for cal in calibrations:
        time_to_cal: float | None = None
        if not cal.is_finished() and cal.start_time is not None:
            time_to_cal = round(cal.start_time - now, 1)

        response.append(
            {
                "name": cal.name,
                "start_time": format_timestamp(cal.start_time),
                "max_start_time": format_timestamp(cal.max_start_time),
                "after": cal.model.after,
                "time_to_cal": time_to_cal,
                "status": cal.state.name.lower(),
                "requires_dome": cal.model.dome,
                "close_dome_after": cal.model.close_dome_after,
            }
        )

    return command.finish(calibrations_sjd=schedule.sjd, calibrations=response)


@calibrations.command(name="reset")
@click.argument("CALIBRATION", type=str, required=False)
@click.option("--include-done", is_flag=True, help="Include done calibrations.")
async def calibrations_reset(
    command: OverwatcherCommand,
    calibration: str | None = None,
    include_done: bool = False,
):
    """Resets the status of a calibration."""

    overwatcher = command.actor.overwatcher

    schedule = overwatcher.calibrations.schedule
    for cal in schedule.calibrations:
        if calibration is not None and cal.name != calibration:
            continue

        if (cal.state == CalibrationState.DONE) and not include_done:
            continue

        cal.state = CalibrationState.WAITING

    overwatcher.calibrations.schedule.update_schedule(reset=True)

    return command.finish()


@overwatcher_cli.group()
def observer(*_):
    """Commands the overwatcher observer."""

    pass


@observer.command(name="status")
async def observer_status(command: OverwatcherCommand):
    """Reports the status of the observer."""

    overwatcher = command.actor.overwatcher
    overwatcher_observer = overwatcher.observer
    gort_observer = overwatcher.gort.observer

    tile = gort_observer._tile

    tile_id: int | None = None
    dither_position: int | None = None
    stage: str | None = None
    standard_no: int | None = None

    if overwatcher_observer.is_observing and gort_observer.is_running() and tile:
        tile_id = tile.tile_id
        dither_position = overwatcher.gort.observer.dither_position

        if standards := overwatcher.gort.observer.standards:
            standard_no = standards.current_standard

        stage = gort_observer.get_running_stage()

    return command.finish(
        observer_status={
            "observing": overwatcher_observer.is_observing,
            "cancelling": overwatcher_observer.is_cancelling,
            "focusing": overwatcher.state.focusing,
            "troubleshooting": overwatcher.state.troubleshooting,
            "tile_id": tile_id,
            "dither_position": dither_position,
            "stage": stage,
            "standard_no": standard_no,
        }
    )


@observer.command()
async def schedule_focus_sweep(command: OverwatcherCommand):
    """Schedules a focus sweep before the next tile."""

    command.actor.overwatcher.observer.force_focus = True

    return command.finish()


@overwatcher_cli.command()
@click.option(
    "--force",
    is_flag=True,
    help="Forces the overwatcher to run the complete "
    "shutdown sequence even if the dome is already closed.",
)
async def shutdown(command: OverwatcherCommand, force: bool = False):
    """Disables the overwatcher, stops observing, and closes the dome."""

    await command.actor.overwatcher.shutdown(force=force)
    return command.finish()


@overwatcher_cli.group()
def transparency():
    """Transparency commands."""

    pass


@transparency.command(name="status")
async def transparency_status(command: OverwatcherCommand):
    """Reports the transparency status of the science telescope."""

    overwatcher = command.actor.overwatcher
    transparency = overwatcher.transparency

    now = time.time()
    if transparency.last_updated < now - 120:
        command.warning("Transparency data is stale.")
        return command.finish(
            transparency={
                "telescope": "sci",
                "mean_zp": None,
                "quality": "unknown",
                "trend": "unknown",
            }
        )

    zp = transparency.zero_point["sci"]

    return command.finish(
        transparency={
            "telescope": "sci",
            "mean_zp": None if math.isnan(zp) else round(zp, 2),
            "quality": transparency.get_quality_string("sci"),
            "trend": transparency.get_trend_string("sci"),
        }
    )


@transparency.command()
async def start_monitoring(command: OverwatcherCommand):
    """Starts monitoring the transparency."""

    overwatcher = command.actor.overwatcher

    if not overwatcher.transparency.is_monitoring():
        await overwatcher.transparency.start_monitoring()
        command.info("Starting transparency monitoring.")

    elapsed: float = 0
    while True:
        if not overwatcher.transparency.is_monitoring():
            return command.finish("Transparency monitoring has been stopped.")

        await asyncio.sleep(1)
        elapsed += 1

        if elapsed >= 30:
            await command.child_command("transparency status")
            elapsed = 0


@transparency.command()
async def stop_monitoring(command: OverwatcherCommand):
    """Stops monitoring the transparency."""

    overwatcher = command.actor.overwatcher

    if overwatcher.transparency.is_monitoring():
        await overwatcher.transparency.stop_monitoring()

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-09
# @Filename: pointing_model.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
from functools import partial

from typing import Sequence

import numpy
import polars
from astropy import units as uu
from astropy.coordinates import (
    AltAz,
    EarthLocation,
    SkyCoord,
    uniform_spherical_random_surface,
)
from astropy.time import Time

from gort import Gort


def get_random_sample(
    n_points: int,
    alt_range: tuple[float, float] | None = None,
    az_range: tuple[float, float] | None = None,
):
    """Provides a random sample of RA/Dec points on the surface of a sphere.

    Parameters
    ----------
    n_points
        Number of points to return. This number is ensured even if ``alt_range``
        or ``az_range`` are provided.
    alt_range
        The range of altitude to which to limit the sample, in degrees.
    az_range
        The range of azimuth to which to limit the sample, in degrees.

    Returns
    -------
    coordinates
        A 2D array with the Alt/Az coordinates of the points on the sphere.

    """

    points = numpy.zeros((0, 2), dtype=numpy.float64)

    lco = EarthLocation.of_site("Las Campanas Observatory")
    now = Time.now()

    while True:
        sph_points = uniform_spherical_random_surface(n_points)
        altaz = AltAz(
            alt=sph_points.lat.deg * uu.deg,
            az=sph_points.lon.deg * uu.deg,
            location=lco,
            obstime=now,
        )

        if alt_range is not None:
            alt = altaz.alt.deg
            altaz = altaz[(alt > alt_range[0]) & (alt < alt_range[1])]
        if az_range is not None:
            az = altaz.az.deg
            altaz = altaz[(az > az_range[0]) & (az < az_range[1])]

        altaz = SkyCoord(altaz[altaz.alt.deg > 30])

        altaz_array = numpy.array([altaz.alt.deg, altaz.az.deg], dtype=numpy.float64).T

        points = numpy.vstack((points, altaz_array))

        if points.shape[0] >= n_points:
            return points[0:n_points, :]


async def get_offset(
    gort: Gort,
    telescope: str,
    ra: float,
    dec: float,
    exposure_time: float = 5,
    add_point: bool = True,
):
    """Determines the offset between a pointing an the measured coordinates.

    Parameters
    ----------
    gort
        The instance of `.Gort` used to command the telescope.
    telescope
        The telescope: ``'sci'``, ``'spec'``, etc.
    ra,dec
        The RA/Dec coordinates of the field to observe.
    exposure_time
        The exposure time.
    add_point
        Whether to add the point to the PWI model.

    Returns
    -------
    offset
        A dictionary with the commanded and measured RA/Dec and the offset,
        or `None` if the measurement failed.

    """

    await gort.telescopes[telescope].goto_coordinates(ra=ra, dec=dec)

    replies = await gort.guiders[telescope].actor.commands.guide(
        reply_callback=partial(gort.guiders[telescope].log_replies, skip_debug=False),
        ra=ra,
        dec=dec,
        exposure_time=exposure_time,
        one=True,
        apply_corrections=False,
    )

    replies = replies.flatten()
    if "measured_pointing" not in replies:
        return None

    return_dict = {"commanded_ra": ra, "commanded_dec": dec, "telescope": telescope}
    return_dict["seqno"] = replies["frame"]["seqno"]
    measured_pointing = replies["measured_pointing"]
    return_dict["measured_ra"] = measured_pointing["ra"]
    return_dict["measured_dec"] = measured_pointing["dec"]
    return_dict["offset_ra"] = measured_pointing["radec_offset"][0]
    return_dict["offset_dec"] = measured_pointing["radec_offset"][1]
    return_dict["offset_ax0"] = measured_pointing["motax_offset"][0]
    return_dict["offset_ax1"] = measured_pointing["motax_offset"][1]
    return_dict["separation"] = measured_pointing["separation"]

    if add_point:
        await gort.telescopes[telescope].actor.commands.modelAddPoint(
            measured_pointing["ra"] / 15,
            measured_pointing["dec"],
        )

    return return_dict


async def pointing_model(
    output_file: str | pathlib.Path | None,
    n_points: int,
    alt_range: tuple[float, float],
    az_range: tuple[float, float],
    telescopes: Sequence[str] = ["sci", "spec", "skye", "skyw"],
    calculate_offset: bool = True,
    home: bool = True,
    add_points: bool = True,
) -> polars.DataFrame | None:
    """Iterates over a series of points on the sky measuring offsets.

    Parameters
    ----------
    output_file
        A parquet file where to save the resulting table.
    n_points
        Number of points on the sky to measure.
    alt_range
        The range of altitude to which to limit the sample, in degrees.
    az_range
        The range of azimuth to which to limit the sample, in degrees.
    telescopes
        The list of telescopes to expose.
    calculate_offset
        Determines the offset between the commanded and observed pointings
        by exposing the AG cameras and solving the field.
    home
        Whether to home the telescope before starting.
    add_points
        Add points to the PWI model. Ignored if ``calculate_offset=False``.

    """

    lco = EarthLocation.of_site("Las Campanas Observatory")

    gort = await Gort(verbosity="debug").init()

    points = get_random_sample(n_points, alt_range=alt_range, az_range=az_range)

    data: polars.DataFrame | None = None

    if output_file is not None:
        outputs_dir = pathlib.Path(__file__).parents[2] / "outputs"
        output_file = pathlib.Path(output_file)
        if not output_file.is_absolute():
            output_file = outputs_dir / output_file
        output_file.parent.mkdir(parents=True, exist_ok=True)

        data = polars.read_parquet(str(output_file))

    assert data is None or isinstance(data, polars.DataFrame)

    if home:
        await gort.telescopes.goto_named_position("zenith")
        await gort.telescopes.home()

    for npoint, (alt, az) in enumerate(points):
        print()

        altaz = SkyCoord(
            AltAz(
                alt=alt * uu.deg,
                az=az * uu.deg,
                obstime=Time.now(),
                location=lco,
            )
        )
        icrs = altaz.transform_to("icrs")
        ra = icrs.ra.deg
        dec = icrs.dec.deg

        gort.log.info(f"({npoint + 1}/{n_points}): Going to {ra:.6f}, {dec:.6f}.")

        if calculate_offset is False:
            tasks = [
                gort.telescopes[telescope].goto_coordinates(ra=ra, dec=dec)
                for telescope in telescopes
            ]
            await asyncio.gather(*tasks)
            await asyncio.sleep(3)

            continue

        results = await asyncio.gather(
            *[
                get_offset(gort, tel, ra, dec, add_point=add_points)
                for tel in telescopes
            ],
            return_exceptions=True,
        )

        valid = []
        for ii, tel in enumerate(telescopes):
            result = results[ii]
            if result is None:
                gort.log.warning(
                    f"Failed determining offset for telescope {tel} "
                    f"at ({ra:.6f}, {dec:.6f})"
                )
            elif isinstance(result, BaseException):
                gort.log.warning(f"Telescope {tel} failed with error: {str(result)}")
            else:
                valid.append(result)

        if len(valid) == 0:
            continue

        this_offsets = polars.DataFrame(valid)

        print()
        print(this_offsets)

        if data is not None:
            data = polars.concat([data, this_offsets])
        else:
            data = this_offsets

        if output_file:
            data.write_parquet(str(output_file))

    return data


if __name__ == "__main__":
    NPOINTS = 50
    ALT_RANGE = (40, 88)
    AZ_RANGE = (0, 355)
    TELESCOPES = ["sci", "spec", "skye", "skyw"]

    OUTPUT = "pointing_model_sci_10points_add.h5"

    asyncio.run(
        pointing_model(
            OUTPUT,
            NPOINTS,
            ALT_RANGE,
            AZ_RANGE,
            telescopes=TELESCOPES,
            home=True,
            add_points=True,
        )
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-04-09
# @Filename: events.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
from datetime import UTC, datetime

from typing import TYPE_CHECKING

from lvmopstools.pubsub import Message, Subscriber

from gort.enums import ErrorCode, Event
from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.tools import add_night_log_comment, decap, insert_to_database


if TYPE_CHECKING:
    from gort.overwatcher.overwatcher import Overwatcher


class MonitorEvents(OverwatcherModuleTask["EventsOverwatcher"]):
    """Processes the notification queue."""

    name = "monitor_events"
    keep_alive = True
    restart_on_error = True

    _running_tasks: list[asyncio.Task] = []

    async def task(self):
        """Runs the task."""

        async for message in Subscriber().iterator(decode=True):
            # Clean done tasks.
            self._running_tasks = [t for t in self._running_tasks if not t.done()]

            task = asyncio.create_task(self.process(message))
            self._running_tasks.append(task)

    async def process(self, message: Message):
        """Processes a notification"""

        message_type = message.message_type

        if message_type != "event":
            return

        event = Event(message.event or Event.UNCATEGORISED)
        name = event.name
        payload = message.payload

        try:
            self.write_to_db(event, payload)
        except Exception as ee:
            self.log.error(f"Failed to write event {name} to the database: {decap(ee)}")

        if event == Event.OBSERVER_NEW_TILE:
            tile_id = payload.get("tile_id", None)
            dither_position = payload.get("dither_position", 0)
            if tile_id is not None:
                await self.overwatcher.notify(
                    f"Observing tile {tile_id} on dither position #{dither_position}."
                )

        elif event == Event.DOME_OPENING:
            await self.overwatcher.notify("The dome is opening ...")

        elif event == Event.DOME_OPEN:
            await self.overwatcher.notify("The dome is now open.")

        elif event == Event.DOME_CLOSING:
            await self.overwatcher.notify("The dome is closing ...")

        elif event == Event.DOME_CLOSED:
            await self.overwatcher.notify("The dome is now closed.")

        elif event == Event.EMERGENCY_SHUTDOWN:
            await self.overwatcher.notify(
                "An emergency shutdown was triggered.",
                level="warning",
            )

        elif event == Event.ERROR:
            await self.handle_error_event(payload)

    def write_to_db(self, event: Event, payload: dict):
        """Writes the event to the database."""

        dt = datetime.now(tz=UTC)

        insert_to_database(
            self.gort.config["services.database.tables.events"],
            [{"date": dt, "event": event.name.upper(), "payload": json.dumps(payload)}],
        )

    async def handle_error_event(self, error_payload: dict):
        """Handles an error event."""

        error = error_payload.get("error", "unespecified error")

        code = error_payload.get("error_code", ErrorCode.UNCATEGORISED_ERROR)
        error_code = ErrorCode(code)

        error_message = f"Error {error_code.value} ({error_code.name}): {decap(error)}"
        if not error_message.endswith("."):
            error_message += "."

        is_observing = self.gort.observer.is_running()
        tile = self.gort.observer._tile
        if is_observing and tile and tile.tile_id:
            await self.notify(
                f"An error event was reported while observing tile {tile.tile_id}. "
                f"{error_message}",
                level="error",
            )
            await add_night_log_comment(
                f"Tile {tile.tile_id} - Error reported. {error_message} "
                "Data quality may have been affected.",
                category="overwatcher",
            )

        else:
            await self.notify(
                f"An error event was reported. {error_message}",
                level="error",
            )


class EventsOverwatcher(OverwatcherModule):
    name = "events"

    tasks = [MonitorEvents()]

    def __init__(self, overwatcher: Overwatcher):
        super().__init__(overwatcher)

        self.queue = asyncio.Queue()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: nps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from gort.devices.core import GortDevice, GortDeviceSet


if TYPE_CHECKING:
    from gort import ActorReply
    from gort.gort import Gort


__all__ = ["NPS", "NPSSet"]


class NPS(GortDevice):
    """Class representing a networked power switch."""

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

    async def status(self, outlet: int | str | None = None):
        """Retrieves the status of the power outlet."""

        if outlet is None:
            reply: ActorReply = await self.actor.commands.status(n_retries=3)
            return reply.flatten()["outlets"]
        else:
            reply: ActorReply = await self.actor.commands.status(outlet, n_retries=3)
            return reply.flatten()["outlet_info"]

    async def on(self, outlet: str):
        """Turns an outlet on."""

        await self.actor.commands.on(outlet, n_retries=3)

    async def off(self, outlet: str):
        """Turns an outlet on."""

        await self.actor.commands.off(outlet, n_retries=3)

    async def all_off(self):
        """Turns off all the outlets."""

        self.write_to_log("Turning off all outlets.")
        await self.actor.commands.all_off(n_retries=3)


class NPSSet(GortDeviceSet[NPS]):
    """A set of networked power switches."""

    __DEVICE_CLASS__ = NPS
    __DEPLOYMENTS__ = ["lvmnps"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-11
# @Filename: transparency.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
from time import time

from typing import TYPE_CHECKING, Literal, Sequence, TypedDict, cast, get_args

import numpy
import polars

from sdsstools.utils import GatheringTaskGroup

from gort.exceptions import GortError
from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.tools import cancel_task, decap, get_lvmapi_route


if TYPE_CHECKING:
    pass


__all__ = ["TransparencyOverwatcher", "TransparencyQuality", "TransparencyQuality"]


class TransparencyQuality(enum.Flag):
    """Flags for transparency status."""

    GOOD = enum.auto()
    POOR = enum.auto()
    BAD = enum.auto()
    UNKNOWN = enum.auto()
    IMPROVING = enum.auto()
    WORSENING = enum.auto()
    FLAT = enum.auto()


Telescopes = Literal["sci", "spec", "skye", "skyw"]


class TransparencyQualityDict(TypedDict):
    sci: TransparencyQuality
    skye: TransparencyQuality
    skyw: TransparencyQuality
    spec: TransparencyQuality


class TransparencyZPDict(TypedDict):
    sci: float
    skye: float
    skyw: float
    spec: float


class TransparencyMonitorTask(OverwatcherModuleTask["TransparencyOverwatcher"]):
    """Monitors transparency."""

    name = "transparency_monitor"
    keep_alive = True
    restart_on_error = True

    def __init__(self):
        super().__init__()

        self.unavailable: bool = False

    async def task(self):
        """Updates the transparency data."""

        n_failures: int = 0

        while True:
            try:
                await self.update_data()
            except Exception as err:
                if not self.unavailable:
                    self.log.error(f"Failed to get transparency data: {decap(err)}")
                n_failures += 1
            else:
                self.module.last_updated = time()
                self.unavailable = False
                n_failures = 0
            finally:
                if n_failures >= 5 and not self.unavailable:
                    await self.notify(
                        "Cannot retrieve transparency data. Will continue trying but "
                        "transparency monitoring will be unavailable.",
                        level="error",
                    )

                    self.module.reset()
                    self.unavailable = True

            await asyncio.sleep(60)

    async def update_data(self):
        """Retrieves and evaluates transparency data."""

        now: float = time()
        lookback: float = 3600

        # Get transparency data from the API for the last hour.
        data = await get_lvmapi_route(
            "/transparency",
            params={"start_time": now - lookback, "end_time": now},
        )

        self.module.data_start_time = data["start_time"]
        self.module.data_end_time = data["end_time"]

        data = (
            polars.DataFrame(
                data["data"],
                orient="row",
                schema={
                    "date": polars.String(),
                    "timestamp": polars.Float64(),
                    "telescope": polars.String(),
                    "zero_point": polars.Float32(),
                },
            )
            .with_columns(
                date=polars.col.date.str.to_datetime(time_zone="UTC", time_unit="ms")
            )
            .sort("telescope", "date")
        )

        # Add a rolling mean.
        data = data.with_columns(
            zero_point_10m=polars.col.zero_point.rolling_mean_by(
                by="date",
                window_size="10m",
            ).over("telescope")
        )

        # Get last 5 and 15 minutes of data.
        data_10 = data.filter(polars.col.timestamp > (now - 300))
        data_15 = data.filter(polars.col.timestamp > (now - 900))

        # Use the last 5 minutes of data to determine the transparency status
        # and value the last 15 minutes to estimate the trend.
        for tel in ["sci", "spec", "skye", "skyw"]:
            data_tel_10 = data_10.filter(polars.col.telescope == tel)
            data_tel_15 = data_15.filter(polars.col.telescope == tel)

            if len(data_tel_10) < 5:
                self.module.quality[tel] = TransparencyQuality.UNKNOWN
                self.module.zero_point[tel] = numpy.nan
                continue

            avg_10 = data_tel_10["zero_point_10m"].median()
            if avg_10 is not None:
                avg_10 = cast(float, avg_10)
                self.module.zero_point[tel] = round(float(avg_10), 2)

                if avg_10 < -22.75:
                    self.module.quality[tel] = TransparencyQuality.GOOD
                elif avg_10 > -22.75 and avg_10 < -22.25:
                    self.module.quality[tel] = TransparencyQuality.POOR
                else:
                    self.module.quality[tel] = TransparencyQuality.BAD

            time_15m = data_tel_15["timestamp"].to_numpy() - data_tel_15["timestamp"][0]
            zp_15m = data_tel_15["zero_point_10m"].to_numpy()
            gradient_15m = (zp_15m[-1] - zp_15m[0]) / (time_15m[-1] - time_15m[0])

            if gradient_15m > 5e-4:
                self.module.quality[tel] |= TransparencyQuality.WORSENING
            elif gradient_15m < -5e-4:
                self.module.quality[tel] |= TransparencyQuality.IMPROVING
            else:
                self.module.quality[tel] |= TransparencyQuality.FLAT

        return data


class TransparencyOverwatcher(OverwatcherModule):
    """Monitors alerts."""

    name = "alerts"

    tasks = [TransparencyMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.last_updated: float = 0

        self.data_start_time: float = 0
        self.data_end_time: float = 0

        self._monitor_task: asyncio.Task | None = None

        self.reset()

    def reset(self):
        """Reset values."""

        self.data_start_time: float = 0
        self.data_end_time: float = 0

        self.quality = TransparencyQualityDict(
            sci=TransparencyQuality.UNKNOWN,
            skye=TransparencyQuality.UNKNOWN,
            skyw=TransparencyQuality.UNKNOWN,
            spec=TransparencyQuality.UNKNOWN,
        )

        self.zero_point = TransparencyZPDict(
            sci=numpy.nan,
            skye=numpy.nan,
            skyw=numpy.nan,
            spec=numpy.nan,
        )

        self.data: polars.DataFrame = polars.DataFrame(
            None,
            schema={
                "time": polars.Datetime(time_unit="ms", time_zone="UTC"),
                "zero_point": polars.Float32(),
                "telescope": polars.String(),
            },
        )

    def write_to_log(
        self,
        telescopes: Sequence[Telescopes] | Telescopes = get_args(Telescopes),
    ):
        """Writes the current state to the log."""

        if isinstance(telescopes, str):
            telescopes = [telescopes]

        for tel in telescopes:
            self.log.info(
                f"Transparency for {tel}: quality={self.get_quality_string(tel)}; "
                f"trend={self.get_trend_string(tel)}; zp={self.zero_point[tel]:.2f}"
            )

    def get_quality_string(self, telescope: Telescopes) -> str:
        """Returns the quality as a string."""

        quality_flag = self.quality[telescope]
        quality: str = "UNKNOWN"

        if quality_flag & TransparencyQuality.BAD:
            quality = "BAD"
        elif quality_flag & TransparencyQuality.POOR:
            quality = "POOR"
        elif quality_flag & TransparencyQuality.GOOD:
            quality = "GOOD"

        return quality

    def get_trend_string(self, telescope: Telescopes) -> str:
        """Returns the trend as a string."""

        quality_flag = self.quality[telescope]
        trend: str = "UNKNOWN"

        if quality_flag & TransparencyQuality.IMPROVING:
            trend = "IMPROVING"
        elif quality_flag & TransparencyQuality.WORSENING:
            trend = "WORSENING"
        elif quality_flag & TransparencyQuality.FLAT:
            trend = "FLAT"

        return trend

    def is_monitoring(self):
        """Returns True if the transparency monitor is running."""

        return self._monitor_task is not None and not self._monitor_task.done()

    async def start_monitoring(self):
        """Starts monitoring transparency."""

        if self.is_monitoring():
            return

        self.gort.log.info("Starting transparency monitor.")
        self._monitor_task = asyncio.create_task(self._monitor_transparency())

    async def stop_monitoring(self):
        """Stops monitoring transparency."""

        if self.is_monitoring():
            self.gort.log.info("Stopping transparency monitor.")
            self._monitor_task = await cancel_task(self._monitor_task)
            await self.gort.guiders.stop()

    async def _monitor_transparency(self):
        """Monitors transparency."""

        # Stop guiding and fold all telescopes except sci.
        await self.gort.guiders.stop()
        async with GatheringTaskGroup() as group:
            for tel in ["spec", "skye", "skyw"]:
                group.create_task(
                    self.gort.telescopes[tel].park(
                        disable=False,
                        kmirror=False,
                    )
                )

        # Start monitoring with the sci telescope.
        sci_task = asyncio.create_task(self.gort.guiders["sci"].monitor(sleep=30))

        while True:
            await asyncio.sleep(30)

            if sci_task.done():
                self.gort.log.error("sci guider has stopped monitoring transparency.")
                await self.stop_monitoring()
                raise GortError("sci guider has stopped monitoring transparency.")

            sci_quality = self.quality["sci"]
            sci_zero_point = self.zero_point["sci"]

            if sci_quality & TransparencyQuality.GOOD:
                self.gort.log.info("sci guider has detected good transparency.")
                await cancel_task(sci_task)
                await self.stop_monitoring()
                break

            self.log.info(
                f"sci guider transparency quality: {sci_quality.name} "
                f"(zero_point={sci_zero_point:.2f})."
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-10-27
# @Filename: alerts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
from time import time

from typing import TYPE_CHECKING

from httpx import ReadTimeout
from lvmopstools.utils import Trigger
from pydantic import BaseModel

from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.tools import decap, get_lvmapi_route


if TYPE_CHECKING:
    pass


__all__ = ["AlertsOverwatcher", "ActiveAlert"]


class AlertsSummary(BaseModel):
    """Summary of alerts."""

    humidity_alert: bool | None = None
    dew_point_alert: bool | None = None
    wind_alert: bool | None = None
    rain: bool | None = None
    door_alert: bool | None = None
    camera_temperature_alert: bool | None = None
    camera_alerts: dict[str, bool] | None = None
    e_stops: bool | None = None
    o2_alert: bool | None = None
    o2_room_alerts: dict[str, bool] | None = None
    heater_alert: bool | None = None
    heater_camera_alerts: dict[str, bool] | None = None


class ConnectivityStatus:
    """Status of the connectivity."""

    def __init__(self):
        self.lco = Trigger(n=3)
        self.internet = Trigger(n=3)


class ActiveAlert(enum.Flag):
    """Flags for active alerts."""

    HUMIDITY = enum.auto()
    DEW_POINT = enum.auto()
    WIND = enum.auto()
    RAIN = enum.auto()
    DOOR = enum.auto()
    CAMERA_TEMPERATURE = enum.auto()
    O2 = enum.auto()
    E_STOPS = enum.auto()
    ALERTS_UNAVAILABLE = enum.auto()
    DISCONNECTED = enum.auto()
    DOME_LOCKED = enum.auto()
    UNKNOWN = enum.auto()

    ALWAYS_CLOSE = HUMIDITY | DEW_POINT | WIND | RAIN
    NO_CLOSE = DOOR | E_STOPS


class AlertsMonitorTask(OverwatcherModuleTask["AlertsOverwatcher"]):
    """Monitors the alerts state."""

    name = "alerts_monitor"
    keep_alive = True
    restart_on_error = True

    INTERVAL: float = 20

    async def task(self):
        """Updates the alerts data."""

        n_failures: int = 0

        while True:
            try:
                await self.update_alerts()
            except Exception as err:
                self.log.error(f"Failed to get alerts data: {decap(err)}")
                n_failures += 1
            else:
                self.module.last_updated = time()
                self.module.unavailable = False
                n_failures = 0
            finally:
                if self.module.unavailable is False and n_failures >= 5:
                    self.module.unavailable = True

            await asyncio.sleep(self.INTERVAL)

    async def update_alerts(self):
        """Processes the weather update and determines whether it is safe to observe."""

        data = await self.module.update_status()

        if data is None:
            raise ValueError("No alerts data available.")

        # For some very critical alerts, we require them to be not null (null here
        # means no data was available or the API failed getting the alert data).
        if data.rain is None or data.humidity_alert is None or data.wind_alert is None:
            raise ValueError("Incomplete alerts data.")


class AlertsOverwatcher(OverwatcherModule):
    """Monitors alerts."""

    name = "alerts"

    tasks = [AlertsMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.state: AlertsSummary | None = None
        self.connectivity = ConnectivityStatus()

        self.last_updated: float = 0.0
        self.unavailable: bool = False

    def is_safe(self) -> tuple[bool, ActiveAlert]:
        """Determines whether it is safe to open."""

        if self.state is None:
            self.log.warning("Alerts data not available. is_safe() returns False.")
            return False, ActiveAlert.UNKNOWN

        is_safe: bool = True
        active_alerts = ActiveAlert(0)

        if self.unavailable:
            self.log.warning("Alerts data is unavailable.")
            active_alerts |= ActiveAlert.ALERTS_UNAVAILABLE
            return False, active_alerts

        if self.unavailable is False and time() - self.last_updated > 300:
            # If the data is not unavailable but it has not been updated
            # in the last 5 minutes, something is wrong. We mark it as unavailable.
            self.log.warning("Alerts data has not been updated in the last 5 minutes.")
            self.unavailable = True
            active_alerts |= ActiveAlert.ALERTS_UNAVAILABLE
            return False, active_alerts

        if self.state.rain:
            self.log.warning("Rain alert detected.")
            active_alerts |= ActiveAlert.RAIN
            is_safe = False

        if self.state.humidity_alert:
            self.log.warning("Humidity alert detected.")
            active_alerts |= ActiveAlert.HUMIDITY
            is_safe = False

        if self.state.dew_point_alert:
            self.log.warning("Dew point alert detected.")
            active_alerts |= ActiveAlert.DEW_POINT
            is_safe = False

        if self.state.wind_alert:
            self.log.warning("Wind alert detected.")
            active_alerts |= ActiveAlert.WIND
            is_safe = False

        if self.state.e_stops:
            self.log.warning("E-stops triggered.")
            active_alerts |= ActiveAlert.E_STOPS
            is_safe = False

        if self.connectivity.internet.is_set():
            self.log.warning("Internet connectivity lost.")
            active_alerts |= ActiveAlert.DISCONNECTED
            is_safe = False

        if self.connectivity.lco.is_set():
            self.log.warning("Internal LCO connectivity lost.")
            active_alerts |= ActiveAlert.DISCONNECTED
            is_safe = False

        if self.overwatcher.dome.locked:
            self.log.warning("Dome is locked.")
            active_alerts |= ActiveAlert.DOME_LOCKED
            is_safe = False

        if self.state.door_alert:
            active_alerts |= ActiveAlert.DOOR
            is_safe = False

        # These alerts are not critical but we log them.
        # TODO: maybe we do want to do something about these alerts.
        if self.state.camera_temperature_alert:
            self.log.warning("Camera temperature alert detected.")
            active_alerts |= ActiveAlert.CAMERA_TEMPERATURE

        if self.state.o2_alert:
            self.log.warning("O2 alert detected.")
            active_alerts |= ActiveAlert.O2

        return is_safe, active_alerts

    async def update_status(self) -> AlertsSummary:
        """Returns the alerts report."""

        alerts_data = await get_lvmapi_route("/alerts/summary")
        summary = AlertsSummary(**alerts_data)

        try:
            summary.e_stops = await self.gort.enclosure.e_stops.status()
        except Exception:
            self.log.warning("Failed to retrieve e-stop status.")

        # For connectivity we want to avoid one single failure to trigger an alert
        # which closes the dome. The connectivity status is a set of triggers that
        # need several settings to be activated.
        try:
            connectivity_data = await get_lvmapi_route(
                "/alerts/connectivity",
                timeout=10,
            )
        except ReadTimeout:
            connectivity_data = {"lco": False, "internet": False}

        if not connectivity_data["lco"]:
            self.connectivity.lco.set()
        else:
            self.connectivity.lco.reset()

        if not connectivity_data["internet"]:
            self.connectivity.internet.set()
        else:
            self.connectivity.internet.reset()

        self.state = summary

        return summary



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-27
# @Filename: observer.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from time import time

from typing import TYPE_CHECKING

from lvmopstools.retrier import Retrier

from gort.exceptions import GortError, OverwatcherError, TroubleshooterTimeoutError
from gort.exposure import Exposure
from gort.overwatcher import OverwatcherModule
from gort.overwatcher.core import OverwatcherModuleTask
from gort.overwatcher.transparency import TransparencyQuality
from gort.tile import Tile
from gort.tools import cancel_task, decap, ensure_period, run_in_executor


if TYPE_CHECKING:
    pass


__all__ = ["ObserverOverwatcher"]


class ObserverMonitorTask(OverwatcherModuleTask["ObserverOverwatcher"]):
    """Monitors the observer status."""

    name = "observer_monitor"
    keep_alive = False
    restart_on_error = True

    interval: float = 1

    async def task(self):
        """Handles whether we should start the observing loop."""

        # These checks will only start the observing loop. If the weather is
        # unsafe or daytime has been reached the main task will handle stopping
        # the loop.

        state = self.overwatcher.state
        ephemeris = self.overwatcher.ephemeris

        while True:
            if state.dry_run:
                pass

            elif self.module.is_observing or self.module.is_cancelling:
                pass

            elif not state.enabled or state.troubleshooting:
                pass

            elif not ephemeris.ephemeris:
                pass

            elif state.calibrating:
                pass

            elif not state.safe:
                pass

            elif ephemeris.is_night(mode="observer"):
                # Start observing if it's night (after evening twilight) but not too
                # close to the morning twilight.
                try:
                    await self.module.start_observing()
                except Exception as err:
                    await self.notify(
                        "An error occurred while starting the "
                        f"observing loop: {decap(err)}",
                        level="error",
                    )
                    await asyncio.sleep(15)

            await asyncio.sleep(self.interval)


class ObserverOverwatcher(OverwatcherModule):
    name = "observer"
    delay = 5

    tasks = [ObserverMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.observe_loop: asyncio.Task | None = None
        self.next_exposure_completes: float = 0

        self.focusing: bool = False
        self._starting_observations: bool = False
        self._cancelling: bool = False

        self.force_focus: bool = False  # Force focus before the next tile

    @property
    def is_observing(self) -> bool:
        """Returns whether the observer is currently observing."""

        if self._starting_observations:
            return True

        return self.observe_loop is not None and not self.observe_loop.done()

    @property
    def is_cancelling(self) -> bool:
        """Returns whether the observer is currently cancelling."""

        if not self.is_observing:
            self._cancelling = False
            return False

        return self._cancelling

    def cancel(self):
        """Requests the cancellation of the observing loop."""

        if self.is_observing and not self.is_cancelling:
            self._cancelling = True
            self.gort.observer.cancelling = True

    async def start_observing(self):
        """Starts observations."""

        if self.overwatcher.state.dry_run:
            return

        if self.is_observing or self.is_cancelling:
            return

        if self.overwatcher.state.calibrating:
            raise GortError("Cannot start observing while calibrating.")

        if not self.overwatcher.state.safe:
            raise GortError("Cannot safely open the telescope.")

        await self.notify("Starting observations.")

        if not (await self.overwatcher.dome.is_opening()):
            try:
                self._starting_observations = True
                await self.notify("Running the start-up sequence and opening the dome.")
                await self.overwatcher.startup()
            except Exception:
                # If the dome failed to open that will disable the
                # Overwatcher and prevent the startup to run again.
                self.log.error("Startup routine failed.")
                return
            finally:
                self._starting_observations = False

        self.observe_loop = asyncio.create_task(self.observe_loop_task())
        await asyncio.sleep(1)

    async def stop_observing(
        self,
        immediate: bool = False,
        reason: str | None = None,
        block: bool = False,
    ):
        """Stops observations."""

        if self.overwatcher.state.dry_run:
            return

        if not self.is_observing:
            return

        # Prepare the notification message.
        msg = "Stopping observations "
        if immediate:
            msg += "immediately."
        else:
            msg += "after the current tile completes."

        if reason is not None:
            msg += f" Reason: {decap(reason)}"

        await self.overwatcher.notify(ensure_period(msg), level="info")

        # Cancel the observing loop.
        self.cancel()

        # If immediate, cancel the task now and cleanup.
        if immediate:
            self.observe_loop = await cancel_task(self.observe_loop)

            # The guiders may have been left running or the spectrograph may still
            # be exposing. Clean up to avoid issues.
            await self.gort.cleanup(readout=False)

        # If block, wait for the observing loop to finish.
        if block and self.observe_loop and not self.observe_loop.done():
            await self.observe_loop
            self.observe_loop = None

    async def observe_loop_task(self):
        """Runs the observing loop."""

        # Check that we have not started too early (this happens when we open the dome
        # a bit early to avoid wasting time). If so, wait until we are properly in
        # the observing window.
        if not self.overwatcher.ephemeris.is_night():
            time_to_twilight = self.overwatcher.ephemeris.time_to_evening_twilight()
            if time_to_twilight and time_to_twilight > 0:
                await self.notify(
                    f"Waiting until evening twilight to start observing "
                    f"({time_to_twilight:.0f} seconds)."
                )
                await asyncio.sleep(time_to_twilight)

        await self.gort.cleanup(readout=True)
        observer = self.gort.observer

        n_tile_positions = 0
        schedule_shudtown: bool = False

        while True:
            try:
                # Wait in case the troubleshooter is doing something.
                await self.overwatcher.troubleshooter.wait_until_ready(300)

                # We want to avoid re-acquiring the tile between dithers. We call
                # the scheduler here and control the dither position loop ourselves.
                tile: Tile = await run_in_executor(Tile.from_scheduler)

                self.log.info(
                    f"Received tile {tile.tile_id} from scheduler: "
                    f"observing dither positions {tile.dither_positions}."
                )

                await self.check_focus(force=n_tile_positions == 0 or self.force_focus)

                for dpos in tile.dither_positions:
                    exp: Exposure | bool = False

                    await self.overwatcher.troubleshooter.wait_until_ready(300)

                    if not self.overwatcher.ephemeris.is_night(mode="observer"):
                        await self.notify(
                            "Twilight will be reached before the next exposure "
                            "completes. Stopping observations now."
                        )
                        self.cancel()
                        schedule_shudtown = True
                        break

                    # The exposure will complete in 900 seconds + acquisition + readout
                    self.next_exposure_completes = time() + 90 + 900 + 60

                    if not (await self.pre_observe_checks()):
                        raise OverwatcherError("Pre-observe checks failed.")

                    result, exps = await observer.observe_tile(
                        tile=tile,
                        dither_position=dpos,
                        async_readout=True,
                        keep_guiding=True,
                        skip_slew_when_acquired=True,
                        run_cleanup=False,
                        cleanup_on_interrupt=True,
                        show_progress=False,
                    )

                    n_tile_positions += 1

                    # Clear counts of errors that are reset
                    # when a tile is successfully observed.
                    self.overwatcher.troubleshooter.reset()

                    if not result and not self.is_cancelling:
                        raise GortError("The observation ended with error state.")

                    if result and len(exps) > 0:
                        exp = exps[0]

                    try:
                        await self.post_exposure(exp)
                    except Exception as err:
                        await self.notify(
                            f"Failed to run post-exposure routine: {decap(err)}",
                            level="error",
                        )

                    if self.is_cancelling:
                        break

            except asyncio.CancelledError:
                break

            except TroubleshooterTimeoutError:
                await self.notify(
                    "The troubleshooter timed out after 300 seconds. "
                    "Cancelling observations.",
                    level="critical",
                )
                break

            except Exception as err:
                if await self.overwatcher.troubleshooter.handle(err):
                    if self.is_cancelling:
                        break
                    if self.focusing:
                        # Force a new focus if the error occurred while focusing.
                        self.force_focus = True

                    continue

            finally:
                self.exposure_completes = 0

                if self.is_cancelling:
                    try:
                        if exp is not False and not exp.done():
                            self.log.warning("Waiting for exposure to read.")
                            await asyncio.wait_for(exp, timeout=80)
                    except Exception:
                        self.log.error("Failed reading last exposure.")

                    break

        await self.gort.cleanup(readout=False)
        await self.notify("The observing loop has ended.")

        if schedule_shudtown:
            # Do not set shutdown_pending until here to prevent the main overwatcher
            # task cancelling the last exposure.
            self.overwatcher.state.shutdown_pending = True

    async def check_focus(self, force: bool = False):
        """Checks if it's time to focus the telescope."""

        should_focus: bool = False

        # Retry focusing up to 2 times with a 10 second delay and a 150 second timeout.
        focus_retrier = Retrier(max_attempts=2, delay=10, timeout=150)

        if not force:
            # Check if we should refocus.
            focus_info = await self.gort.guiders.sci.get_focus_info()
            focus_age = focus_info["reference_focus"]["age"]

            if focus_age is None or focus_age > 2 * 3600:  # Focus sweeps every 2 hours
                should_focus = True

        else:
            should_focus = True

        # Focus when the loop starts or every 1 hour or at the beginning
        # of the loop.
        if should_focus:
            try:
                self.focusing = True
                await self.notify("Focusing telescopes.")
                await focus_retrier(self.gort.guiders.focus)()
            except Exception as err:
                await self.notify(
                    f"Failed while focusing the telescopes twice: {decap(err)}",
                    level="error",
                )
                await self.notify(
                    "I will continue observations and try to "
                    "focus again after the next tile."
                )
                self.force_focus = True
            else:
                self.force_focus = False
            finally:
                self.focusing = False

    async def pre_observe_checks(self):
        """Runs pre-observe checks."""

        if await self.gort.specs.are_errored():
            exp = self.gort.specs.last_exposure
            if exp and not exp.done():
                self.log.warning(
                    "Spectrographs are idle but an exposure is ongoing. "
                    "Waiting for it to finish before resetting."
                )
                try:
                    await asyncio.wait_for(exp, timeout=100)
                except asyncio.TimeoutError:
                    self.log.error(
                        "Timed out waiting for exposure to finish. "
                        "Resetting spectrographs."
                    )

            await self.gort.specs.reset()

        return True

    async def post_exposure(self, exp: Exposure | bool):
        """Runs post-exposure checks."""

        if self._cancelling:
            return

        if exp is False:
            raise GortError("No exposure was returned.")

        # Output transparency data for the last exposure.
        transparency = self.overwatcher.transparency
        transparency.write_to_log(["sci"])

        # TODO: for now if the transparency is bad we close and disable the overwatcher.
        if transparency.quality["sci"] & TransparencyQuality.BAD:
            await self.overwatcher.shutdown(
                reason="Transparency is bad. Stopping observations and closing the "
                "dome. Resume observations manually when the transparency is deemed "
                "good.",
                level="warning",
                close_dome=True,
                disable_overwatcher=True,
            )

        return

        # TODO: monitor transparency and resume observing automatically.
        if transparency.quality["sci"] & TransparencyQuality.BAD:
            await self.notify(
                "Transparency is bad. Stopping observations and starting "
                "the transparency monitor.",
            )

            # If we reach twilight this will cause the overwatcher
            # to immediately stop observations.
            self.exposure_completes = 0

            try:
                await asyncio.wait_for(
                    transparency.start_monitoring(),
                    timeout=3600,
                )

            except asyncio.TimeoutError:
                await self.notify("Transparency monitor timed out.", level="warning")
                await self.overwatcher.shutdown(
                    reason="Transparency has been bad for over one hour.",
                    disable_overwatcher=True,
                )

            else:
                # The transparency monitor has ended. There are two possible reasons:

                # - Something stopped the observing loop and with it the monitor.
                #   Do nothing and return. The main task will handle the rest.
                if self._cancelling:
                    return

                # - The transparency is good and the monitor has ended.
                if transparency.quality["sci"] & TransparencyQuality.GOOD:
                    await self.notify("Transparency is good. Resuming observations.")
                    return

                else:
                    await self.notify(
                        "Transparency is still bad but the monitor stopped. "
                        "Triggering shutdown.",
                    )
                    await self.overwatcher.shutdown(
                        reason="Transparency monitor failed.",
                        disable_overwatcher=True,
                    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-11
# @Filename: exposure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import pathlib
import re
import warnings
from collections import defaultdict

from typing import TYPE_CHECKING, Any, Awaitable, Callable, Literal, Sequence

from astropy.io import fits
from astropy.time import Time
from rich.progress import BarColumn, MofNCompleteColumn, Progress, TextColumn

from sdsstools.time import get_sjd

from gort.exceptions import ErrorCode, GortSpecError
from gort.tools import (
    cancel_task,
    decap,
    get_md5sum,
    get_md5sum_from_spectro,
    insert_to_database,
    is_interactive,
    is_notebook,
    run_in_executor,
)


if TYPE_CHECKING:
    from gort.gort import Gort


__all__ = ["Exposure", "READOUT_TIME"]


READOUT_TIME = 55

HOOKS_TYPE = defaultdict[
    Literal["pre-readout", "post-readout"],
    list[Callable[[Any], Awaitable]],
]


class Exposure(asyncio.Future["Exposure"]):
    """A class representing an exposure taken by a :obj:`.SpectrographSet`.

    Parameters
    ----------
    gort
        A `.Gort` instance to communicate with the actors.
    exp_no
        The exposure sequence number. If :obj:`None`, the next valid
        sequence number will be used.
    flavour
        The image type. Defaults to ``'object'``.
    object
        The object name to be added to the header.
    specs
        List the spectrographs to expose. Defaults to all.

    Attributes
    ----------
    hooks
        A dictionary of hooks to call in specific steps of the exposure. Each
        hook must be a list of coroutines to call in that specific situation.
        All coroutines for a given hook are called concurrently and depending
        on the hook they may be scheduled as a task and not awaited. Available
        hooks are:
        - ``'pre-readout'`` which is called with the header before readout
        begins; the coroutine can modify the header in place or perform
        any tasks that should be complete at the end of integration.
        - ``'post-readout'`` called as a task (not awaited) after the readout
        is complete. Receives the :obj:`.Exposure` object.

        To add a coroutine to a hook ::

            async def update_header(header):
                header.update({'KEY': 1})

            exp = Exposure(g)
            exp.hooks['pre-readout'].append(update_header)

    """

    def __init__(
        self,
        gort: Gort,
        exp_no: int | None = None,
        flavour: str | None = "object",
        object: str | None = "",
        specs: Sequence[str] | None = None,
    ):
        self.gort = gort
        self.specs = gort.specs
        self.devices = specs
        self.exp_no = exp_no or self.specs.get_expno()
        self.flavour = flavour or "object"
        self.object = object or ""

        self.start_time = Time.now()
        self._exposure_time: float | None = None

        self.error: bool = False
        self.reading: bool = False

        self._timer_task: asyncio.Task | None = None
        self._progress: Progress | None = None

        self.hooks: HOOKS_TYPE = defaultdict(
            list,
            {"pre-readout": [], "post-readout": []},
        )

        if self.flavour not in ["arc", "object", "flat", "bias", "dark"]:
            raise GortSpecError(
                "Invalid flavour type.",
                error_code=ErrorCode.USAGE_ERROR,
            )

        super().__init__()

        self.add_done_callback(self._when_done)

    def __repr__(self):
        return (
            f"<Exposure (exp_no={self.exp_no}, flavour={self.flavour}, "
            f"object={self.object!r} error={self.error}, reading={self.reading}, "
            f"done={self.done()})>"
        )

    async def expose(
        self,
        exposure_time: float | None = None,
        header: dict | None = None,
        async_readout: bool = False,
        show_progress: bool | None = None,
        object: str | None = None,
        raise_on_error: bool = True,
    ):
        """Exposes the spectrograph.

        Parameters
        ----------
        exposure_time
            The exposure time.
        header
            A dictionary with the extra header values..
        async_readout
            Returns after integration completes. Readout is initiated
            but handled asynchronously and can be await by awaiting
            the returned :obj:`.Exposure` object.
        show_progress
            Displays a progress bar with the elapsed exposure time.
            If :obj:`None` (the default), will show the progress bar only
            in interactive sessions.
        object
            The object name to be passed to the header.
        raise_on_error
            Whether to raise an error when the exposure is marked as errored.

        """

        log = self.specs.write_to_log

        if self.specs.last_exposure is not None and not self.specs.last_exposure.done():
            log("Waiting for previous exposure to read out.", "warning")
            await self.specs.last_exposure

        # Check that all specs are idle and not errored.
        status = await self.specs.status(simple=True)
        for spec_name, spec_status in status.items():
            if self.devices is not None and spec_name not in self.devices:
                continue

            if "IDLE" not in spec_status["status_names"]:
                raise GortSpecError(
                    "Some spectrographs are not IDLE.",
                    error_code=ErrorCode.SECTROGRAPH_NOT_IDLE,
                )
            if "ERROR" in spec_status["status_names"]:
                raise GortSpecError(
                    "Some spectrographs have ERROR status. "
                    "Solve this manually before exposing.",
                    error_code=ErrorCode.SECTROGRAPH_NOT_IDLE,
                )

        await self.specs.reset()

        header = header or {}

        # Set object name for header.
        if "OBJECT" not in header:
            if object is not None:
                header.update({"OBJECT": object})
            elif self.object is not None and self.object != "":
                header.update({"OBJECT": self.object})
            elif self.flavour != "object":
                header.update({"OBJECT": self.flavour})
            else:
                header.update({"OBJECT": ""})

        if show_progress is None:
            show_progress = is_interactive() or is_notebook()

        if exposure_time is None and self.flavour != "bias":
            raise GortSpecError(
                "Exposure time required for all flavours except bias.",
                error_code=ErrorCode.USAGE_ERROR,
            )

        warnings.filterwarnings("ignore", message=".*cannot modify a done command.*")

        self.specs.last_exposure = self

        monitor_task: asyncio.Task | None = None

        self._exposure_time = exposure_time or 0.0

        log_msg = f"Starting integration on spectrograph exposure {self.exp_no} "
        if self.flavour == "bias":
            log_msg += f"({self.flavour})."
        else:
            log_msg += f"({self.flavour or object}, {exposure_time:.1f} s)."
        log(log_msg, "info")

        try:
            if show_progress:
                await self.start_timer(self._exposure_time)

            self.start_time = Time.now()

            await self.specs.send_command_all(
                "expose",
                devices=self.devices,
                flavour=self.flavour,
                exposure_time=exposure_time,
                seqno=self.exp_no,
                readout=False,
                timeout=(exposure_time or 0) + 30,
            )

            # Call pre-readout tasks.
            await self._call_hook("pre-readout", header)

            # At this point we have integrated and are ready to read.
            self.reading = True
            log(f"Reading spectrograph exposure {self.exp_no} ...", "info")
            readout_task = asyncio.create_task(
                self.specs.send_command_all(
                    "read",
                    devices=self.devices,
                    header=json.dumps(header),
                    timeout=90,
                )
            )

            # Now launch the task that marks the Future done when the spec
            # is IDLE. If async_readout=False then that will return immediately
            # because the spec is already idle. If async_readout=True, this method
            # will return now and the task will mark the Future done when readout
            # complete (readout is ongoing and does not need to be launched).
            monitor_task = asyncio.create_task(self._done_monitor())
            if not async_readout:
                await readout_task
                await monitor_task
            else:
                self.stop_timer()

        except Exception as err:
            # Cancel the monitor task
            await cancel_task(monitor_task)

            self.error = True
            self.set_result(self)

            if raise_on_error:
                raise GortSpecError(
                    f"Exposure failed with error {err}",
                    error_code=ErrorCode.SECTROGRAPH_FAILED_EXPOSING,
                )
            else:
                log.warning(f"Exposure failed with error {err}")

        finally:
            self.stop_timer()

        return self

    async def start_timer(
        self,
        exposure_time: float,
        readout_time: float = READOUT_TIME,
    ):
        """Starts the rich timer."""

        self.stop_timer()

        self._progress = Progress(
            TextColumn(f"[yellow]({self.exp_no})"),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=None),
            MofNCompleteColumn(),
            TextColumn("s"),
            expand=True,
            transient=True,
            auto_refresh=False,
            refresh_per_second=1,
            console=self.specs.gort._console,  # Need to use same console as logger.
        )

        exp_task = self._progress.add_task(
            "[blue] Integrating ...",
            total=int(exposure_time),
        )

        readout_task = self._progress.add_task(
            "[blue] Reading ...",
            total=int(readout_time),
            visible=False,
        )

        self._progress.start()

        async def update_timer():
            elapsed = 0
            while True:
                if self._progress is None:
                    return
                elif elapsed < exposure_time:
                    self._progress.update(exp_task, advance=1)
                else:
                    self._progress.update(
                        exp_task,
                        description="[green] Integration complete",
                        completed=int(exposure_time),
                    )
                    self._progress.update(readout_task, advance=1, visible=True)

                asyncio.create_task(run_in_executor(self._progress.refresh))

                await asyncio.sleep(1)
                elapsed += 1

        self._timer_task = asyncio.create_task(update_timer())

        return

    def stop_timer(self):
        """Cancels the timer."""

        if self._timer_task and not self._timer_task.done():
            self._timer_task.cancel()
        self._timer_task = None

        if self._progress:
            for task in self._progress.task_ids:
                self._progress.remove_task(task)
            self._progress.stop()
            self._progress.console.clear_live()
        self._progress = None

    async def verify_files(self):
        """Checks that the files have been written and have the right contents."""

        config = self.specs.gort.config["specs"]

        HEADERS_CRITICAL = config["verification"]["headers"]["critical"]
        HEADERS_WARNING = config["verification"]["headers"]["warning"]

        for retry in range(3):
            # Very infrequently we have cases in which this check fails even if the
            # files are there. It may be a delay in the NFS disk or a race condition
            # somewhere. As a hotfile, we try three times with a delay before giving up.

            files = self.get_files()

            n_spec = len(self.devices) if self.devices else len(config["devices"])
            n_files_expected = n_spec * 3

            if (n_files_found := len(files)) < n_files_expected:
                if retry < 2:
                    await asyncio.sleep(3)
                    continue

                self.specs.write_to_log(
                    f"Expected {n_files_expected} files but found {n_files_found}."
                    "warning"
                )

        for file in files:
            header = fits.getheader(str(file))

            for key in HEADERS_CRITICAL:
                if key not in header:
                    raise RuntimeError(f"Keyword {key} not present in {file!s}")

            for key in HEADERS_WARNING:
                if key not in header:
                    self.specs.write_to_log(
                        f"Keyword {key} not present in {file!s}",
                        "warning",
                    )

            md5sum_spectro = get_md5sum_from_spectro(file)
            md5sum = get_md5sum(file)
            if md5sum != md5sum_spectro:
                raise RuntimeError(f"MD5 checksum validation failed for file {file!r}")

        if len(files) > 0:
            pattern_path = re.sub("([rbz][1-3])", "*", str(files[0]))
            self.specs.write_to_log(f"Files saved to {pattern_path!r}.")

        return files

    def get_files(self):
        """Returns the files written by the exposure."""

        sjd = get_sjd("LCO")
        config = self.specs.gort.config
        data_path = pathlib.Path(config["specs"]["data_path"].format(SJD=sjd))

        return list(data_path.glob(f"*-[0]*{self.exp_no}.fits.gz"))

    def _when_done(self, result):
        """Called when the future is done."""

        asyncio.create_task(self._call_hook("post-readout", self, as_task=True))

    async def _done_monitor(self):
        """Waits until the spectrographs are idle, and marks the Future done."""

        log = self.specs.write_to_log

        await self.specs.send_command_all("wait_until_idle", allow_errored=True)

        self.reading = False

        for spec in self.specs.values():
            reply = await spec.status(simple=True)
            if "ERROR" in reply["status_names"]:
                self.error = True

        log(f"Spectrograph exposure {self.exp_no} has completed.", "info")

        files = await self.verify_files()
        exposure_table = self.gort.config["services.database.tables.exposures"]

        try:
            async with asyncio.timeout(10):
                await self.write_to_db(files, exposure_table)
        except asyncio.TimeoutError:
            log(
                "Timeout writing to database. Data may not have been recorded.",
                "warning",
            )
        except Exception as err:
            log(
                f"Error writing to database: {decap(err)}. "
                "Data may not have been recorded.",
                "warning",
            )

        # Set the Future.
        self.set_result(self)

    @staticmethod
    async def write_to_db(files: list[pathlib.Path], table_name: str):
        """Records the exposures in ``gortdb.exposure``."""

        column_data: list[dict[str, Any]] = []

        for file in files:
            header_ap = dict(await run_in_executor(fits.getheader, str(file)))
            header_ap.pop("COMMENT", None)

            header = {kk.upper(): vv for kk, vv in header_ap.items() if vv is not None}

            exposure_no = header.get("EXPOSURE", None)
            image_type = header.get("IMAGETYP", None)
            exposure_time = header.get("EXPTIME", None)
            spec = header.get("SPEC", None)
            ccd = header.get("CCD", None)
            mjd = header.get("SMJD", None)
            start_time = header.get("INTSTART", None)
            tile_id = header.get("TILE_ID", None)

            if start_time is not None:
                start_time = Time(start_time, format="isot").datetime

            column_data.append(
                {
                    "exposure_no": exposure_no,
                    "image_type": image_type,
                    "exposure_time": exposure_time,
                    "spec": spec,
                    "ccd": ccd,
                    "mjd": mjd,
                    "start_time": start_time,
                    "tile_id": tile_id,
                    "header": json.dumps(header),
                }
            )

        await run_in_executor(insert_to_database, table_name, column_data)

    async def _call_hook(self, hook_name: str, *args, as_task: bool = False, **kwargs):
        """Calls the coroutines associated with a hook."""

        if hook_name not in self.hooks:
            raise ValueError(f"Invalid hook {hook_name!r}.")

        self.specs.write_to_log(f"Calling hook {hook_name!r}.")

        coros = self.hooks[hook_name]
        if not isinstance(coros, list) and asyncio.iscoroutinefunction(coros):
            task = asyncio.create_task(coros(*args, **kwargs))
        elif isinstance(coros, list):
            task = asyncio.gather(*[coro(*args, **kwargs) for coro in coros])
        else:
            raise ValueError(f"Invalid hook functions found for {hook_name!r}.")

        if not as_task:
            await task



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-02-07
# @Filename: remote.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from types import SimpleNamespace

from typing import TYPE_CHECKING, Any, Callable, Literal

import unclick
from aiormq import AMQPConnectionError, ChannelInvalidStateError
from lvmopstools.retrier import Retrier
from typing_extensions import Self

from clu.tools import CommandStatus

from gort.exceptions import (
    GortTimeoutError,
    GortWarning,
    InvalidRemoteCommand,
    RemoteCommandError,
)
from gort.tools import get_valid_variable_name


if TYPE_CHECKING:
    from clu.client import AMQPReply
    from clu.command import Command

    from gort.devices.core import GortDevice
    from gort.gort import Gort


__all__ = ["RemoteActor", "RemoteCommand", "ActorReply"]


class CommandSet(dict[str, "RemoteCommand"]):
    """A command set for a remote actor."""

    def __getattribute__(self, __name: str) -> RemoteCommand:
        if __name in self:
            return self[__name]
        return super().__getattribute__(__name)


class RemoteActor:
    """A programmatic representation of a remote actor."""

    def __init__(self, client: Gort, name: str, device: GortDevice | None = None):
        self.client = client

        self.name = name
        self.model: dict = {}
        self.commands = CommandSet()
        self.device = device

    def __repr__(self):
        return f"<RemoteActor (name={self.name})>"

    async def init(self) -> Self:
        """Initialises the representation of the actor."""

        if (
            self.client.connection.connection
            and self.client.connection.connection.is_closed
        ):
            raise RuntimeError("Gort is not connected.")

        cmd = await self.send_raw_command("get-command-model")
        if cmd.status.did_fail:
            warnings.warn(f"Cannot get model for actor {self.name}.", GortWarning)
            return self

        self.model = cmd.replies.get("command_model")

        commands_dict = {}
        for command_info in self.model["commands"].values():
            command_name = get_valid_variable_name(command_info["name"])
            commands_dict[command_name] = RemoteCommand(self, command_info)

        self.commands = CommandSet(commands_dict)

        return self

    async def send_raw_command(self, *args, **kwargs):
        """Sends a raw command to the actor.

        The parameters are the same as CLU's ``AMQPClient.send_command()`` with
        the exception of the consumer name, which is replaced with the current actor
        name.

        """

        # This lock prevents that if the client has disconnected we get multiple
        # reconnections.
        async with self.client._connect_lock:
            try:
                cmd = await self.client.send_command(
                    self.name,
                    *args,
                    await_command=False,
                    **kwargs,
                )
            except (AMQPConnectionError, ChannelInvalidStateError):
                # Client has disconnected. This should only happen if running Gort
                # in an ipython terminal where the event loop only runs while a command
                # is executing. See https://tinyurl.com/4kcwxzx9

                self.client.log.warning("Client has disconnected. Reconnecting.")

                await self.client.start()

                cmd = await self.client.send_command(
                    self.name,
                    *args,
                    await_command=False,
                    **kwargs,
                )

        return await cmd

    async def refresh(self):
        """Refreshes the command list."""

        await self.init()


class RemoteCommand:
    """Representation of a remote command."""

    def __init__(
        self,
        remote_actor: RemoteActor,
        model: dict,
        parent: RemoteCommand | None = None,
    ):
        self._remote_actor = remote_actor
        self._model = model
        self._parent = parent

        self._name = model["name"]
        self.commands = SimpleNamespace()

        self.is_group = "commands" in model and len(model["commands"]) > 0
        if self.is_group:
            for command_info in model["commands"].values():
                command_name = get_valid_variable_name(command_info["name"])
                child_command = RemoteCommand(remote_actor, command_info, parent=self)
                setattr(self.commands, command_name, child_command)

    def get_command_string(self, *args, **kwargs):
        """Gets the command string for a set of arguments."""

        return unclick.build_command_string(self._model, *args, **kwargs)

    async def __call__(
        self,
        *args,
        reply_callback: Callable[[AMQPReply], None] | None | Literal[False] = None,
        timeout: float | None = None,
        n_retries: int = 0,
        delay: float = 1,
        **kwargs,
    ):
        """Executes the remote command with some given arguments, allowing retries.

        If a command fails with a timeout or the command cannot be parsed or does
        not exist, and error is raised immediately without retries.

        Parameters
        ----------
        args,kwargs
            Arguments to pass to :obj:`.get_command_string`.
        reply_callback
            Function to call with each reply received from the command.
        timeout
            Maximum time allowed for the command to complete.
        n_retries
            The maximum number of attempts before giving up.
        delay
            The delay between attempts, in seconds. This delay is increased for the
            second and successive retries using an exponential backoff.

        """

        retrier = Retrier(
            max_attempts=n_retries,
            delay=delay,
            on_retry=self._log_command_retry,
            raise_on_exception_class=[GortTimeoutError, InvalidRemoteCommand],
        )

        return await retrier(self.run)(
            *args,
            reply_callback=reply_callback,
            timeout=timeout,
            **kwargs,
        )

    async def run(
        self,
        *args,
        reply_callback: Callable[[AMQPReply], None] | None | Literal[False] = None,
        timeout: float | None = None,
        **kwargs,
    ):
        """Build the remote command and run it.

        This method does not allow retries. Use :meth:`.__call__` for that, which
        should be generally preferred.

        """

        parent_string = ""
        if self._parent is not None:
            # Call parent chain without arguments. This is not bullet-proof for all
            # cases, but probably good enough for now.
            parent_string = self._parent.get_command_string() + " "

        if reply_callback is None and self._remote_actor.device is not None:
            reply_callback = self._remote_actor.device.log_replies

        cmd_string = parent_string + self.get_command_string(*args, **kwargs)
        cmd = await self._remote_actor.send_raw_command(
            cmd_string,
            callback=reply_callback,
            time_limit=timeout,
        )

        actor_reply = ActorReply(self._remote_actor, cmd)
        for reply in cmd.replies:
            if len(reply.body) > 0:
                actor_reply.replies.append(reply.body)

        actor = self._remote_actor.name
        command_name = self._name

        if cmd.status & CommandStatus.TIMEDOUT:
            raise GortTimeoutError(
                f"Actor {actor!r} timed out executing command {command_name!r}.",
                command=cmd,
                remote_command=self,
            )

        if not cmd.status.did_succeed:
            error = actor_reply.get("error")
            if error and "does not exist or cannot be parsed" in error:
                raise InvalidRemoteCommand(
                    f"Command '{actor} {cmd_string}' does not "
                    "exist or cannot be parsed.",
                    command=cmd,
                    remote_command=self,
                )

            raise RemoteCommandError(
                f"Actor {actor!r} failed executing command {command_name!r}.",
                command=cmd,
                remote_command=self,
                reply=actor_reply,
            )

        return actor_reply

    def _log_command_retry(self, error: BaseException):
        """Logs a command retry."""

        actor = self._remote_actor.name
        command_name = self._name

        if not isinstance(error, RemoteCommandError) or error.reply is None:
            # This should not happen, but just in case.
            self._remote_actor.client.log.warning(
                f"Actor {actor!r} failed executing command {command_name!r}. Retrying.",
            )
            return

        reply_error = error.reply.get("error")
        error_message = f" {reply_error!s}" if reply_error is not None else ""
        if not error_message.endswith("."):
            error_message += "."

        self._remote_actor.client.log.warning(
            f"Actor {actor!r} failed executing command "
            f"{command_name!r}.{error_message}. Retrying.",
        )


@dataclass
class ActorReply:
    """A reply to an actor command.

    Attributes
    ----------
    actor
        The :obj:`.RemoteActor` that generated the reply.
    command
        The CLU command instance.
    replies
        A list of replies generated by the command.

    """

    actor: RemoteActor
    command: Command
    replies: list[dict] = field(default_factory=list)

    def flatten(self):
        """Returns a flattened dictionary of replies.

        Note that if a keyword has been output multiple times, only the
        last value is retained.

        """

        result = {}
        for reply in self.replies:
            for key in reply:
                result[key] = reply[key]

        return result

    def get(self, key: str, default: Any = ...):
        """Returns the first occurrence of a keyword in the reply list."""

        for reply in self.replies:
            if key in reply:
                return reply[key]

        if default is not ...:
            return default

        raise KeyError(f"Keyword {key!r} not found in replies.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-22
# @Filename: test.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from .base import BaseRecipe


__all__ = ["TestRecipe"]


class TestRecipe(BaseRecipe):
    """A test recipe."""

    name = "test_recipe"

    async def recipe(self, fail: bool = False):
        """The test recipe."""

        self.gort.log.info("Running test recipe.")

        await asyncio.sleep(5)

        if fail:
            self.gort.log.error("Test recipe failed.")
            raise ValueError("Test recipe failed.")

        self.gort.log.info("Test recipe completed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-11-01
# @Filename: safety.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from time import time

from typing import ClassVar

from lvmopstools.retrier import Retrier

from gort.overwatcher.alerts import ActiveAlert
from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.tools import decap


__all__ = ["SafetyOverwatcher"]


class SafetyMonitorTask(OverwatcherModuleTask["SafetyOverwatcher"]):
    """Monitors the safety status."""

    name = "safety_monitor"
    keep_alive = False
    restart_on_error = True

    INTERVAL: ClassVar[float] = 30

    def __init__(self):
        super().__init__()

        self.unsafe_since: float | None = None
        self.failed: bool = False

        self.n_get_data_failures: int = 0

    async def task(self):
        """Checks safety conditions and closes the dome after a delay."""

        while True:
            try:
                is_safe, dome_open = await self.get_data()
                self.n_get_data_failures = 0
            except Exception as err:
                if self.n_get_data_failures == 0:
                    self.overwatcher.log.error(
                        f"Error getting safety data: {decap(err)}",
                        exc_info=err,
                    )
                elif self.n_get_data_failures >= 3:
                    await self.overwatcher.notify(
                        "The safety monitor task has failed to get safety data "
                        "three times in a row. The task will stop.",
                        level="critical",
                    )
                    break

                self.n_get_data_failures += 1
                await asyncio.sleep(self.INTERVAL)
                continue

            try:
                if not self.failed and not is_safe and dome_open:
                    # Conditions are unsafe and the dome is open. We give the
                    # main task 5 minutes to close the dome itself. If it fails,
                    # we close the dome ourselves.

                    now = time()

                    if self.unsafe_since is None:
                        self.unsafe_since = now
                    elif now - self.unsafe_since > 5 * 60:
                        await self.overwatcher.notify(
                            "Safety alerts detected and the dome remains open. "
                            "Closing the dome.",
                            level="error",
                        )
                        await self.module.close_dome()

                        # Now run a shutdown. This should not try to close the dome
                        # since that's already done, but it will stop the observe loop,
                        # clean-up, etc.
                        await self.overwatcher.shutdown(
                            reason="safety alerts detected",
                            disable_overwatcher=True,
                        )

                elif self.failed:
                    # We have failed closing the dome as a last resort. We have issued
                    # a critical alert. We don't try closing the dome again.
                    pass

                else:
                    # Conditions are safe or the dome is closed. Reset variables.
                    self.unsafe_since = None
                    self.failed = False

            except Exception as err:
                await self.overwatcher.notify(
                    "The safety monitor task has failed to close the dome. "
                    "Unsafe conditions have been detected. "
                    "Please close the dome immediately!!!",
                    level="critical",
                )

                # Record the task error in the log.
                self.overwatcher.log.critical(
                    f"Error in safety monitor task: {decap(err)}",
                    exc_info=err,
                )

                self.failed = True

            await asyncio.sleep(self.INTERVAL)

    @Retrier(max_attempts=3, delay=5)
    async def get_data(self):
        """Returns the safety status and whether the dome is open."""

        is_safe, status = self.overwatcher.alerts.is_safe()
        dome_open = await self.overwatcher.gort.enclosure.is_open()

        # If is_safe=False but the e-stops are in, we do not want to close the dome.
        if not is_safe and status & ActiveAlert.E_STOPS:
            is_safe = True

        return is_safe, dome_open


class SafetyOverwatcher(OverwatcherModule):
    """Monitors alerts."""

    name = "safety"
    tasks = [SafetyMonitorTask()]
    delay = 10

    @Retrier(max_attempts=3, delay=5)
    async def close_dome(self):
        """Closes the dome. Bypasses all safety checks."""

        # Closes the dome at an extremely low level without parking the telescopes,
        # checking the local status, issuing notifications, etc. This is a last
        # resort to close the dome in case of an emergency and we want to avoid
        # as many steps that could fail as possible.

        await self.gort.enclosure.actor.commands.dome.commands.close(force=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: nps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from datetime import UTC, datetime, timedelta
from functools import partial

from typing import TYPE_CHECKING

import polars
from packaging.version import Version

from gort import config
from gort.devices.core import GortDevice, GortDeviceSet
from gort.enums import GuiderStatus
from gort.exceptions import ErrorCode, GortError, GortGuiderError
from gort.tools import GuiderMonitor, cancel_task


if TYPE_CHECKING:
    from clu import AMQPReply

    from gort.gort import Gort
    from gort.remote import ActorReply


__all__ = ["Guider", "GuiderSet"]


class Guider(GortDevice):
    """Class representing a guider."""

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

        self.separation: float | None = None
        self.status: GuiderStatus = GuiderStatus.IDLE

        self.guider_monitor = GuiderMonitor(self.gort, self.name)
        self.gort.add_reply_callback(self._status_cb)

    @property
    def ag(self):
        """Gets the :obj:`.AG` device associated with this guider."""

        return self.gort.ags[self.name]

    @property
    def telescope(self):
        """Gets the :obj:`.Telescope` device associated with this guider."""

        return self.gort.telescopes[self.name]

    async def update_status(self):
        """Returns the guider status."""

        status_reply = await self.actor.commands.status()

        return status_reply.flatten()

    async def _status_cb(self, reply: AMQPReply):
        """Listens to guider keywords and updates the internal state."""

        if reply.sender == self.actor.name:
            if "status" in reply.body:
                self.status = GuiderStatus(int(reply.body["status"], 16))
            if "measured_pointing" in reply.body:
                self.separation = reply.body["measured_pointing"]["separation"]

    async def wait_until_guiding(
        self,
        guide_tolerance: float | None = None,
        timeout: float | None = None,
    ) -> tuple[bool, GuiderStatus, float | None, bool]:
        """Waits until the guider has converged.

        Parameters
        ----------
        guide_tolerance
            The minimum separation, in arcsec, between the measured and desired
            positions that needs to be reached before returning. If :obj:`None`,
            waits until guiding (as opposed to acquisition) begins.
        timeout
            Maximum time, in seconds, to wait before returning. If :obj:`None`,
            waits indefinitely. If the timeout is reached it does not
            raise an exception.

        Returns
        -------
        reached
            Whether the desired minimum separation was reached.
        status
            The current :obj:`.GS`.
        separation
            The current separation.
        timedout
            :obj:`True` if the acquisition timed out.

        """

        # Initial delay to allow time for the guider to switch to DRIFTING status.
        await asyncio.sleep(1)

        elapsed = 1
        while True:
            has_acquired = (
                self.status is not None
                and self.separation is not None
                and self.status & GuiderStatus.GUIDING
                and not self.status & GuiderStatus.DRIFTING
                and (guide_tolerance is None or self.separation < guide_tolerance)
            )
            if has_acquired:
                return (True, self.status, self.separation, False)

            elapsed += 1
            if timeout is not None and elapsed > timeout:
                return (False, self.status, self.separation, True)

            await asyncio.sleep(1)

    async def expose(self, *args, continuous: bool = False, **kwargs):
        """Exposes this telescope cameras.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to the guider expose command.
        continuous
            Whether to expose the camera continuously. If :obj:`False`
            it takes a single exposure.

        """

        while True:
            await self.actor.commands.expose(*args, **kwargs)

            if not continuous:
                return

    async def focus(
        self,
        inplace=False,
        sweep: bool = True,
        guess: float | None = None,
        step_size: float = config["guiders.focus.step_size"],
        steps: int = config["guiders.focus.steps"],
        exposure_time: float = config["guiders.focus.exposure_time"],
    ):
        """Focus the telescope.

        Parameters
        ----------
        inplace
            If :obj:`True`, focuses the telescope where it is pointing at. Otherwise
            points to zenith.
        sweep
            Performs a focus sweep around the initial guess position to find the
            best focus. If :obj:`False`, the focus position is determined based on
            the current bench temperature.
        guess
            The initial guess for the focuser position. If :obj:`None`, the initial
            guess is determined based on the current bench temperature.
        step_size
            The size, in focuser units, of each step.
        steps
            The total number of step points. Must be an odd number.
        exposure_time
            The exposure time for each step.

        """

        reply_callback = partial(self.log_replies, skip_debug=False)

        if sweep is False:
            self.write_to_log("Adjusting focus position.", "info")
            await self.actor.commands.adjust_focus(reply_callback=reply_callback)
            return

        await self.update_status()

        if self.status & GuiderStatus.NON_IDLE:
            self.write_to_log(
                "Guider is not idle. Stopping it before focusing.",
                level="warning",
            )
            await self.stop()

        # Send telescopes to zenith.
        if not inplace:
            self.write_to_log("Moving telescope to zenith.")
            await self.gort.telescopes[self.name].goto_named_position(
                "zenith",
                altaz_tracking=True,
            )

        try:
            self.write_to_log(f"Focusing telescope {self.name}.", "info")

            replies = await self.actor.commands.focus(
                reply_callback=reply_callback,
                guess=guess,
                step_size=step_size,
                steps=steps,
                exposure_time=exposure_time,
            )

            best_focus = replies.get("best_focus")

            if best_focus is None:
                raise GortError("best_focus keyword was not emitted.")
            elif best_focus["focus"] < 0.5 or best_focus["r2"] < 0.5:
                raise GortError(
                    "Estimated focus does not seem to be correct. "
                    "Please repeat the focus sweep."
                )

            focus = best_focus["focus"]
            fwhm = best_focus["fwhm"]
            self.write_to_log(
                f"Best focus: {fwhm} arcsec at {focus} DT",
                "info",
            )

            return focus, fwhm

        except GortError as err:
            self.write_to_log(f"Failed focusing with error: {err}", level="error")

        finally:
            self.separation = None

        return -999, -999

    async def adjust_focus(self):
        """Adjusts the focus position based on the current bench temperature."""

        await self.focus(sweep=False)

    async def guide(
        self,
        ra: float | None = None,
        dec: float | None = None,
        exposure_time: float = 5.0,
        pixel: tuple[float, float] | str | None = None,
        monitor: bool = True,
        output_monitor_data: bool = True,
        **guide_kwargs,
    ):
        """Starts the guide loop.

        This command blocks until `.stop` is called.

        Parameters
        ----------
        ra,dec
            The coordinates to acquire. If :obj:`None`, the current telescope
            coordinates are used.
        exposure_time
            The exposure time of the AG integrations.
        pixel
            The pixel on the master frame on which to guide. Defaults to
            the central pixel. This can also be the name of a known pixel
            position for this telescope, e.g., ``'P1-1'`` for ``spec``.
        monitor
            Whether to monitor the guide loop and output the average and last
            guide metrics every 30 seconds.
        output_monitor_data
            Whether to output the monitor data to the log.
        guide_kwargs
            Other keyword arguments to pass to ``lvmguider guide``. The includes
            the ``pa`` argument that if not provided is assumed to be zero.

        """

        monitor_task: asyncio.Task | None = None

        # The PA argument in lvmguider was added in 0.4.0a0.
        if self.version == Version("0.99.0") or self.version < Version("0.4.0a0"):
            guide_kwargs.pop("pa")

        self.separation = None

        await self.update_status()

        if self.status & GuiderStatus.NON_IDLE:
            raise GortGuiderError(
                "Guider is not IDLE",
                error_code=ErrorCode.COMMAND_FAILED,
            )

        if ra is None or dec is None:
            status = await self.telescope.status()
            ra_status = status["ra_j2000_hours"] * 15
            dec_status = status["dec_j2000_degs"]

            ra = ra if ra is not None else ra_status
            dec = dec if dec is not None else dec_status

        config = self.gort.config
        if isinstance(pixel, str):
            if pixel not in config["guiders"]["devices"][self.name]["named_pixels"]:
                raise GortGuiderError(
                    f"Invalid pixel name {pixel!r}.",
                    error_code=ErrorCode.INVALID_PIXEL_NAME,
                )
            pixel = config["guiders"]["devices"][self.name]["named_pixels"][pixel]

        log_msg = f"Guiding at RA={ra:.6f}, Dec={dec:.6f}"
        if pixel is not None:
            log_msg += f", pixel=({pixel[0]:.1f}, {pixel[1]:.1f})."
        self.write_to_log(log_msg, level="info")

        try:
            if monitor:
                self.guider_monitor.start_monitoring()
                if output_monitor_data:
                    monitor_task = asyncio.create_task(self._monitor_task())

            await self.actor.commands.guide(
                reply_callback=partial(self.log_replies, skip_debug=False),
                ra=ra,
                dec=dec,
                exposure_time=exposure_time,
                reference_pixel=pixel,
                **guide_kwargs,
            )

        except Exception as err:
            # Deal with the guide command being cancelled when we stop it.
            if "This command has been cancelled" not in str(err):
                raise

        finally:
            await cancel_task(monitor_task)

    async def _monitor_task(self, timeout: float = 30):
        """Monitors guiding and reports average and last guide metrics."""

        while True:
            try:
                await asyncio.sleep(timeout)

                # Get updated data
                df = self.guider_monitor.get_dataframe()
                if df is None:
                    continue

                df = df.filter(polars.col.telescope == self.name)

                # Select columns.
                df = df.select(
                    [
                        "frameno",
                        "time",
                        "n_sources",
                        "focus_position",
                        "fwhm",
                        "ra",
                        "dec",
                        "ra_offset",
                        "dec_offset",
                        "separation",
                        "mode",
                    ]
                )

                # Remove NaN rows.
                df = df.drop_nulls()

                now = datetime.now(UTC)
                time_range = now - timedelta(seconds=timeout)
                time_data = df.filter(polars.col.time > time_range).sort("time")

                if (
                    len(time_data) == 0
                    or "fwhm" not in time_data
                    or "separation" not in time_data
                ):
                    continue

                # Calculate and report last.
                last = time_data.tail(1)
                sep_last = round(last["separation"][0], 3)
                fwhm_last = round(last["fwhm"][0], 2)
                mode_last = last["mode"][0]
                self.write_to_log(
                    f"Last: sep={sep_last} arcsec; fwhm={fwhm_last} arcsec; "
                    f"mode={mode_last!r}",
                    "info",
                )

                # Calculate and report averages.
                sep_avg = round(time_data["separation"].mean(), 3)  # type: ignore
                fwhm_avg = round(time_data["fwhm"].mean(), 2)  # type: ignore
                self.write_to_log(
                    f"Average ({timeout} s): sep={sep_avg} arcsec; "
                    f"fwhm={fwhm_avg} arcsec",
                    "info",
                )

            except asyncio.CancelledError:
                return

            except Exception as err:
                self.write_to_log(f"Error in guider monitor: {err}", "warning")

    async def stop(self) -> None:
        """Stops the guide loop.

        Parameters
        ----------
        wait_until_stopped
            Blocks until the guider is idle.

        """

        self.write_to_log("Stopping guider.", "info")

        await self.actor.commands.stop()
        self.status = GuiderStatus.IDLE
        self.guider_monitor.stop_monitoring()

    async def set_pixel(self, pixel: tuple[float, float] | str | None = None):
        """Sets the master frame pixel on which to guide.

        Parameters
        ----------
        pixel
            The pixel on the master frame on which to guide. Defaults to
            the central pixel. This can also be the name of a known pixel
            position for this telescope, e.g., ``'P1-1'`` for ``spec``.

        """

        config = self.gort.config
        if isinstance(pixel, str):
            if pixel not in config["guiders"]["devices"][self.name]["named_pixels"]:
                raise GortGuiderError(
                    f"Invalid pixel name {pixel!r}.",
                    error_code=ErrorCode.INVALID_PIXEL_NAME,
                )
            pixel = config["guiders"]["devices"][self.name]["named_pixels"][pixel]

        if pixel is None:
            await self.actor.commands.reset_pixel()
        else:
            await self.actor.commands.set_pixel(*pixel)

    async def apply_corrections(self, enable: bool = True):
        """Enable/disable corrections being applied to the axes."""

        await self.actor.commands.corrections(mode="enable" if enable else "disable")

    async def monitor(
        self,
        ra: float | None = None,
        dec: float | None = None,
        exposure_time: float = 5.0,
        sleep: float = 60,
        monitor: bool = True,
    ):
        """Guides at a given position, sleeping between exposures.

        This is a convenience function mainly to monitor transparency during bad
        weather conditions. The telescope will be slewed to a given position
        (default to zenith) and guide with a low cadence. This results in the
        guider keywords, including transparency and FWHM, being output and the
        plots in Grafana being updated.

        After cancelling the monitoring make sure to stop the guiders with the
        :obj:`.Guider.stop` method.

        Parameter
        ---------
        ra,dec
            The coordinates to acquire. If :obj:`None`, the current zenith
            coordinates are used.
        exposure_time
            The exposure time of the AG integrations.
        sleep
            The time to sleep between exposures (seconds).
        monitor
            Start the guider monitor task. Data collected during the monitoring
            can be access as a :obj:`polars.DataFrame` from
            :obj:`Guider.guider_monitor.get_dataframe() <.GuiderMonitor.get_dataframe>`.

        """

        if ra is None and dec is None:
            await self.telescope.goto_named_position("zenith", altaz_tracking=True)

            # Get approximate RA/Dec. It doesn't really matter, we just want to guide
            # on a field that's close to zenith.
            tel_status = await self.telescope.status()
            ra = tel_status.get("ra_j2000_hours")
            dec = tel_status.get("dec_j2000_degs")

            if ra is None or dec is None:
                raise GortGuiderError("Cannot determine telescope RA/Dec.")
            ra *= 15.0

        elif (ra is None and dec is not None) or (ra is not None and dec is None):
            raise ValueError("Both RA and Dec need to be provided.")

        # Even if we already went to zenith in alt/az we need to go to these
        # coordinates again to make sure the kmirror is set.
        await self.telescope.goto_coordinates(ra, dec)

        await self.guide(
            ra=ra,
            dec=dec,
            exposure_time=exposure_time,
            sleep=sleep,
            monitor=monitor,
            output_monitor_data=False,
        )

    async def get_focus_info(self):
        """Returns the guider focus information."""

        info_reply: ActorReply = await self.actor.commands.focus_info()
        info = info_reply.flatten()

        for key in ["reference_focus", "current_focus"]:
            focus_ts = info[key]["timestamp"]

            if focus_ts is None:
                continue

            time = focus_ts.split("T")[1]
            if not focus_ts.endswith("Z") and "+" not in time and "-" not in time:
                focus_ts += "Z"

            info[key]["timestamp"] = datetime.fromisoformat(focus_ts)

        now = info["current_focus"]["timestamp"]
        ref_ts = info["reference_focus"]["timestamp"]

        if now is not None and ref_ts is not None:
            age = (now - ref_ts).total_seconds()
        else:
            age = None

        info["reference_focus"]["age"] = age

        return info


class GuiderSet(GortDeviceSet[Guider]):
    """A set of telescope guiders."""

    __DEVICE_CLASS__ = Guider
    __DEPLOYMENTS__ = ["lvmguider"]

    async def expose(self, *args, continuous: bool = False, **kwargs):
        """Exposes all the cameras using the guider.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to :obj:`.Guider.expose`.
        continuous
            Whether to expose the camera continuously. If :obj:`False`
            it takes a single exposure.

        """

        await self.call_device_method(
            Guider.expose,
            *args,
            continuous=continuous,
            **kwargs,
        )

    async def take_darks(self):
        """Takes AG darks."""

        # Move telescopes to park to prevent light, since we don't have shutters.
        # We use goto_named_position to prevent disabling the telescope and having
        # to rehome.
        self.write_to_log("Moving telescopes to park position.", level="info")
        await self.gort.telescopes.goto_named_position("park")

        # Take darks.
        self.write_to_log("Taking darks.", level="info")

        cmds = []
        for guider in self.values():
            cmds.append(
                guider.actor.commands.expose(
                    flavour="dark",
                    reply_callback=partial(guider.log_replies, skip_debug=False),
                )
            )

        if len(cmds) > 0:
            await asyncio.gather(*cmds)

    async def focus(
        self,
        inplace=False,
        guess: float | dict[str, float] | None = None,
        step_size: float = 0.5,
        steps: int = 7,
        exposure_time: float = 5.0,
    ):
        """Focus all the telescopes.

        Parameters
        ----------
        inplace
            If :obj:`True`, focuses the telescopes where they are pointing at. Otherwise
            points to zenith.
        guess
            The initial guesses for focuser position. If :obj:`None`, an estimate
            based on the current bench temperatures is used. It can also be a float
            value, which will be used for all telescopes, or a mapping of telescope
            name to guess value. Missing values will default to the configuration
            value.
        step_size
            The size, in focuser units, of each step.
        steps
            The total number of step points. Must be an odd number.
        exposure_time
            The exposure time for each step.

        """

        self.write_to_log("Running focus sequence.", "info")

        if guess is None:
            guess_dict = {}
        elif isinstance(guess, dict):
            guess_dict = guess
        else:
            guess_dict = {guider_name: guess for guider_name in self}

        jobs = [
            self[guider_name].focus(
                inplace=inplace,
                sweep=True,
                guess=guess_dict.get(guider_name, None),
                step_size=step_size,
                steps=steps,
                exposure_time=exposure_time,
            )
            for guider_name in self
        ]
        results = await asyncio.gather(*jobs)

        best_focus: list[str] = []
        error: bool = False
        for itel, name in enumerate(self):
            result = results[itel]
            if result is None:
                continue

            best_focus.append(f"{name}: {result[1]}")

            if any(result) < 0:
                error = True

        self.write_to_log("Best focus: " + ", ".join(best_focus), "info")

        if error:
            self.write_to_log("One or more focus values are invalid.", "error")
            return False

        return True

    async def adjust_focus(self):
        """Adjusts the focus position based on the current bench temperature."""

        await asyncio.gather(*[self[gname].adjust_focus() for gname in self])

    async def guide(self, *args, **kwargs):
        """Guide on all telescopes.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to :obj:`.Guider.guide`.

        """

        await self.call_device_method(Guider.guide, *args, **kwargs)

    async def stop(self):
        """Stops the guide loop on all telescopes."""

        await self.call_device_method(Guider.stop)

    async def apply_corrections(self, enable: bool = True):
        """Enable/disable corrections being applied to the axes."""

        await self.call_device_method(Guider.apply_corrections, enable=enable)

    async def monitor(self, *args, **kwargs):
        """Guides at a given position, sleeping between exposures.

        See :obj:`.Guider.monitor` for details.

        """

        await self.call_device_method(Guider.monitor, *args, **kwargs)

    async def wait_until_guiding(
        self,
        names: list[str] | None = None,
        guide_tolerance: float | None = None,
        timeout: float | None = None,
    ):
        """Waits until the guiders have converged.

        Parameters
        ----------
        names
            List of telescopes to wait for convergence.
        guide_tolerance
            The minimum separation, in arcsec, between the measured and desired
            positions that needs to be reached before returning. If :obj:`None`,
            waits until guiding (as opposed to acquisition) begins.
        timeout
            Maximum time, in seconds, to wait before returning. If :obj:`None`,
            waits indefinitely. If the timeout is reached it does not
            raise an exception.

        Returns
        -------
        status
            A dictionary with the telescope names and a tuple indicating whether
            the desired minimum separation was reached. The current
            ``GuiderStatus``, and the current separation for that telescope.

        """

        names = names or list(self)
        results = await asyncio.gather(
            *[
                self[name].wait_until_guiding(
                    guide_tolerance=guide_tolerance,
                    timeout=timeout,
                )
                for name in names
            ]
        )

        return dict(zip(names, results))
